import os
import pickle
import subprocess
from pathlib import Path, PurePath
from subprocess import DEVNULL
from tempfile import mkstemp
from typing import List, Optional, Tuple, Union

import ms3
import pandas as pd
from joblib import Parallel, delayed
from music21.converter import parse
from music21.stream import Measure, Part, Score
from pandas import DataFrame
from tqdm import tqdm

import musif.extract.constants as C
from musif.cache import (
    CACHE_FILE_EXTENSION,
    FileCacheIntoRAM,
    SmartModuleCache,
    store_score_df,
)
from musif.common._constants import GENERAL_FAMILY
from musif.common.exceptions import FeatureError, ParseFileError
from musif.config import ExtractConfiguration
from musif.extract.common import _filter_parts_data
from musif.extract.utils import (
    cast_mixed_dtypes,
    extract_global_time_signature,
    process_musescore_file,
)
from musif.logs import ldebug, lerr, linfo, lwarn, pdebug, perr, pinfo
from musif.musescore import constants as mscore_c
from musif.musicxml import constants as musicxml_c
from musif.musicxml import extract_numeric_tempo, name_parts, split_layers
from musif.musicxml.scoring import (
    _extract_abbreviated_part,
    extract_sound,
    to_abbreviation,
)

_cache = FileCacheIntoRAM(10000)  # To cache scanned scores


def parse_filename(
    file_path: str,
    split_keywords: List[str],
    expand_repeats: bool = False,
    export_dfs_to: Union[str, PurePath] = None,
    remove_unpitched_objects: bool = True,
) -> Score:
    """
    This function parses a musicxml file and returns a music21 Score object. If
    the file has already been parsed, it will be loaded from cache instead of
    processing it again. Split a part in different parts if the instrument
    family is in keywords argument and expands repeats if indicated.

    Parameters
    ----------
    file_path: str
    A path to a music xml path.
    split_keywords: List[str]
     A lists of keywords based on music21 instrument sound names to split in different parts.
    expand_repeats: bool
     Determines whether to expand or not the repetitions. Default value is False.
    export_dfs_to: Union[str, PurePath]
     Path to a directory where dataframes containing the score data are exported. If
     None, no score is exported. Default value is None.
    Returns
    -------
    resp : Score
     The score saved in cache or the new score parsed with the necessary parts split.
    Raises
    ------
     ParseFileError
       If the xml file can't be parsed for any reason.
    """
    score = _cache.get(file_path)
    if score is not None:
        return score
    try:
        score = parse(file_path)
        if export_dfs_to is not None:
            dest_path = Path(export_dfs_to)
            dest_path /= Path(file_path).with_suffix(".pkl").name
            store_score_df(score, dest_path)
        # give a name to all parts in the score
        name_parts(score)
        if remove_unpitched_objects:
            unpitched_objs = list(
                score.flatten().getElementsByClass(["PercussionChord", "Unpitched"])
            )
            score.remove(unpitched_objs, recurse=True)
        split_layers(score, split_keywords)
        if expand_repeats:
            score = score.expandRepeats()
        _cache.put(file_path, score)
    except Exception as e:
        print(file_path)
        raise ParseFileError(file_path) from e
    return score


def parse_musescore_file(file_path: str, expand_repeats: bool = False) -> pd.DataFrame:
    """
    This function parses a musescore file and returns a pandas dataframe. If the file
    has already been parsed, it will be loaded from cache instead of processing it
    again.

    Parameters
    ----------
    file_path: str
        A path to a music mscx path.
    expand_repeats: bool
        Determines whether to expand or not the repetitions. Default value is False.
    Returns
    -------
    resp : pd.DataFrame
        The score saved in cache or the new score parsed in the form of a dataframe.
    Raises
    ------
    ParseFileError
        If the musescore file can't be parsed for any reason.
    """
    harmonic_analysis = _cache.get(file_path)
    if harmonic_analysis is not None:
        return harmonic_analysis
    try:
        harmonic_analysis = process_musescore_file(file_path, expand_repeats)
        _cache.put(file_path, harmonic_analysis)
    except Exception as e:
        harmonic_analysis = None
        print(file_path)
        raise ParseFileError(file_path) from e
    return harmonic_analysis


def find_files(
    extensions: str or List[str],
    base_dir: Union[str, List[Union[str, PurePath]]],
    limit_files: List[str] = None,
    exclude_files: List[str] = None,
) -> List[PurePath]:
    """Extracts the paths to files given an extension

    Given a directory path, return a list of paths of files found, in alphabetic order.
    It searches recursively inside `base_dir`. If `base_dir` is a fileor a list of paths
    or directories with `extension`, it is returned in a list. If given neither a string
    nor list of strings raise a TypeError and if the file doesn't exists returns a
    ValueError.

    Parameters
    ----------
    extension: str or Iterable[str]
        A list of strings representing the extensions that will be looked for
    base_dir : Union[str, Iterable[str]]
        A path or directory
    limit_files: Iterable[str] = None
        List of file names relative to `base_dir`. Only these files are taken.
        Incompatible with `exclude_files`
    exclude_files: Iterable[str] = None
        List of file names relative to `base_dir`. None of these files are taken.
        Incompatible with `limit_files`

    Returns
    -------
    resp : List[PurePath]
      The list of musicxml files found in the provided arguments
      This list will be returned in alphabetical order

    Raises
    ------
    TypeError
      If the type is not the expected (str or List[str]).

    ValueError
      If the provided string is neither a directory nor a file path
    """
    if isinstance(extensions, str):
        extensions = [extensions]

    if base_dir is None:
        return []
    base_dir = Path(base_dir)
    if not base_dir.exists():
        raise ValueError(f"File {base_dir} doesn't exist")
    elif base_dir.is_dir():
        ret = []
        for ext in extensions:
            ret += sorted([f for f in base_dir.glob(f"**/*{ext}") if f.is_file()])
        if limit_files is not None:
            limit_stems = set(map(lambda x: Path(x).stem, limit_files))
            return [f for f in ret if f.stem in limit_stems]
        elif exclude_files is not None:
            exclude_stems = set(map(lambda x: Path(x).stem, exclude_files))
            return [f for f in ret if f.stem not in exclude_stems]
        else:
            return ret
    elif base_dir.is_file() and base_dir.suffix in extensions:
        return [base_dir]
    else:
        return []


#  sorted(obj.glob(f"*{extension}"))
class FeaturesExtractor:
    """
    Extract features for a score or a list of scores, according to the parameters
    established in the configuration files. It extracts musical features using music21
    and ms3 library, based on the configuration and stores them in a dictionary
    (score features) that at the end will be returned as a DataFrame by the
    `extract` method.

    During the parsing, unpitched objects, (e.g. objects referred to percussion
    instruments) may be removed (see the option
    `remove_unpitched_objects` in the configuration).

    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        *args:  Could be a path to a .yml file, an AbstractExtractConfiguration object or a dictionary. Length zero or one.
        **kwargs: Get keywords to construct ExtractConfiguration.
        limit_files: List[str] = None
            List of file names relative to `obj`. Only these files are taken.
            Incompatible with `exclude_files`
        exclude_files: List[str] = None
            List of file names relative to `obj`. None of these files are taken.
            Incompatible with `limit_files`

        Raises
        ------
        TypeError
         - If the type is not the expected (str, dict or ExtractConfiguration).
        ValueError
          - If there is too many arguments(args)
        FileNotFoundError
          - If any of the files/directories path inside the expected configuration doesn't exit.
        """

        self._cfg = ExtractConfiguration(*args, **kwargs)
        self.limit_files = kwargs.get("limit_files") or getattr(
            self._cfg, "limit_files", None
        )
        self.exclude_files = kwargs.get("exclude_files") or getattr(
            self._cfg, "exclude_files", None
        )
        # self.regex = re.compile("from {FEATURES_MODULES}.([\w\.]+) import")
        # creates the directory for the cache
        if self._cfg.cache_dir is not None:
            pinfo("Cache activated!")
            Path(self._cfg.cache_dir).mkdir(exist_ok=True)

        if "jsymbolic" in self._cfg.features:
            from musif.extract.features import jsymbolic

            jsymbolic.utils.download_jsymbolic()
            jsymbolic.utils.get_java_path()

    def extract(self) -> DataFrame:
        """
        Extracts features given in the configuration data getting a file, directory or several file paths,
        returning a DataFrame containing musical features.

        Returns
        ------
        Score dataframe with the extracted features of given scores. For one score only, a DataFrem is returned with one row only.

        Raises
        ------
        ParseFileError
           If the musicxml file can't be parsed for any reason.
        KeyError
           If features aren't loaded in corrected order or dependencies
        """
        linfo("--- Analyzing scores ---\n".center(120, " "))

        xml_filenames = find_files(
            musicxml_c.MUSIC21_FILE_EXTENSIONS,
            self._cfg.xml_dir,
            limit_files=self.limit_files,
            exclude_files=self.exclude_files,
        )
        musescore_filenames = find_files(
            mscore_c.MUSESCORE_FILE_EXTENSION,
            self._cfg.musescore_dir,
            limit_files=self.limit_files,
            exclude_files=self.exclude_files,
        )
        if len(musescore_filenames) == 0:
            if self._cfg.is_requested_musescore_file():
                perr(
                    f"\nMusescore files are needed for the following features {C.REQUIRE_MSCORE}, but cannot find musescore files. Those features won't be computed!"
                )
        if len(xml_filenames) > 0:
            filenames = xml_filenames
        elif len(musescore_filenames) > 0:
            filenames = musescore_filenames
        elif self._cfg.cache_dir is not None:
            filenames = find_files(
                CACHE_FILE_EXTENSION,
                self._cfg.cache_dir,
                limit_files=self.limit_files,
                exclude_files=self.exclude_files,
            )
        else:
            filenames = []
        if len(filenames) == 0:
            raise FileNotFoundError("No file found for extracting features!")

        score_df = self._process_corpus(filenames)

        # fix dtypes
        score_df = score_df.convert_dtypes()
        score_df = score_df.apply(cast_mixed_dtypes, axis=0)

        return score_df

    def _process_corpus(
        self, filenames: List[PurePath]
    ) -> Tuple[List[dict], List[dict]]:
        def process_corpus_par(idx, filename):
            try:
                if self._cfg.window_size is not None:
                    score_features = self._process_score_windows(idx, filename)
                else:
                    score_features = self._process_score(idx, filename)
            except Exception as e:
                if self._cfg.ignore_errors:
                    lerr(
                        f"Error while extracting features for file {filename}, skipping it because `ignore_errors` is True!"
                    )
                    return {}
                else:
                    raise e
            return score_features

        scores_features = Parallel(n_jobs=self._cfg.parallel)(
            delayed(process_corpus_par)(idx, fname)
            for idx, fname in enumerate(tqdm(filenames))
        )

        if self._cfg.window_size is not None:
            all_dfs = []
            for score in scores_features:
                df_score = DataFrame(score)
                df_score = df_score.reindex(sorted(df_score.columns), axis=1)
                df_score.replace("NA", pd.NA, inplace=True)
                all_dfs.append(df_score)
            all_dfs = pd.concat(all_dfs, axis=0, keys=range(len(all_dfs)))
        else:
            all_dfs = DataFrame(scores_features)
            all_dfs = all_dfs.reindex(sorted(all_dfs.columns), axis=1)
            all_dfs = all_dfs.replace("NA", pd.NA)
        return all_dfs

    def _init_score_processing(self, idx: int, filename: PurePath):
        if self._cfg.cache_dir is not None:
            cache_name = (
                Path(self._cfg.cache_dir)
                / filename.parent
                / (filename.name + CACHE_FILE_EXTENSION)
            )
            cache_name.parent.mkdir(parents=True, exist_ok=True)
        else:
            cache_name = None
        score_data = self._get_score_data(filename, load_cache=cache_name)
        parts_data = [
            self._get_part_data(score_data, part)
            for part in score_data[C.DATA_SCORE].parts
        ]
        parts_data = _filter_parts_data(parts_data, self._cfg.parts_filter)
        basic_features = self.extract_modules(
            self._cfg.basic_modules_addresses, score_data, parts_data, basic=True
        )
        basic_features[C.ID] = idx
        return basic_features, cache_name, parts_data, score_data

    def _process_score(self, idx: int, filename: PurePath) -> dict:

        (
            basic_features,
            cache_name,
            parts_data,
            score_data,
        ) = self._init_score_processing(idx, filename)
        extract_global_time_signature(score_data)
        score_features = self.extract_modules(
            self._cfg.feature_modules_addresses, score_data, parts_data, basic=False
        )
        score_features = {**basic_features, **score_features}
        score_features[C.WINDOW_ID] = 0

        if self._cfg.cache_dir is not None:
            pickle.dump(score_data, open(cache_name, "wb"))
        return score_features

    def _process_score_windows(self, idx: int, filename: PurePath) -> List[dict]:
        (
            basic_features,
            cache_name,
            parts_data,
            score_data,
        ) = self._init_score_processing(idx, filename)

        extract_global_time_signature(score_data)

        window_features = {}
        nmeasures = len(score_data[C.DATA_SCORE].parts[0].getElementsByClass(Measure))

        ws = self._cfg.window_size
        hopsize = ws - self._cfg.overlap
        number_windows = (nmeasures - self._cfg.overlap) // hopsize

        all_windows_features = []
        for idx in range(number_windows):
            first_window_measure = idx * hopsize
            last_window_measure = first_window_measure + ws
            window_data, window_parts_data = self._select_window_data(
                score_data, parts_data, first_window_measure, last_window_measure
            )
            window_data.update(
                {k: v for k, v in score_data.items() if k not in window_data}
            )

            window_features = self.extract_modules(
                self._cfg.feature_modules_addresses,
                window_data,
                window_parts_data,
                basic=False,
            )

            window_features[
                C.WINDOW_RANGE
            ] = f"{first_window_measure} - {last_window_measure}"

            window_features[C.WINDOW_ID] = idx

            window_features = {**basic_features, **window_features}

            all_windows_features.append(window_features)
            first_window_measure = last_window_measure - self._cfg.overlap

        if self._cfg.cache_dir is not None:
            pickle.dump(score_data, open(cache_name, "wb"))
        return all_windows_features

    def _select_window_data(
        self, score_data: dict, parts_data: list, first_measure: int, last_measure: int
    ):
        window_score = score_data[C.DATA_SCORE].measures(
            first_measure, last_measure, indicesNotNumbers=True
        )
        filtered_partNames = [i.partName for i in score_data["parts"]]
        window_parts = [
            i for i in window_score.parts if i.partName in filtered_partNames
        ]
        if (
            self._cfg.is_requested_musescore_file()
            and score_data[C.DATA_MUSESCORE_SCORE] is not None
        ):
            window_mscore = score_data[C.DATA_MUSESCORE_SCORE].loc[
                (score_data[C.DATA_MUSESCORE_SCORE]["mn"] <= last_measure)
                & (score_data[C.DATA_MUSESCORE_SCORE]["mn"] >= first_measure)
            ]
            window_mscore.reset_index(inplace=True, drop=True, level=0)
        else:
            window_mscore = None
        window_score_data = {
            C.DATA_SCORE: window_score,
            C.DATA_FILTERED_PARTS: window_parts,
            C.DATA_MUSESCORE_SCORE: window_mscore,
            C.DATA_NUMERIC_TEMPO: score_data[C.DATA_NUMERIC_TEMPO],
        }

        for i, p in enumerate(window_parts):
            parts_data[i]["part"] = p
        return window_score_data, parts_data

    def extract_modules(
        self, packages: list, data: dict, parts_data: dict, basic: bool
    ):
        score_features = {}
        parts_features = [{} for _ in range(len(parts_data))]
        for package in packages:
            for module in self._find_modules(package, basic):
                self._update_parts_module_features(
                    module, data, parts_data, parts_features
                )
                self._update_score_module_features(
                    module, data, parts_data, parts_features, score_features
                )
        return score_features

    def _load_xml_data(self, filename: Union[str, PurePath]):
        filename = Path(filename)
        if filename.suffix == mscore_c.MUSESCORE_FILE_EXTENSION:
            # convert to xml in a temporary file
            mscore = self._cfg.mscore_exec
            if mscore is None:
                mscore = ms3.utils.get_musescore("auto")
            if mscore is None:
                raise RuntimeError(
                    "Cannot find musescore executable. Please provide xml files or the path to a musescore installation with the configuration `mscore_exec`"
                )
            if not isinstance(mscore, (list, tuple)):
                # this is needed to allow stuffs like `xvfb-run -a mscore`
                mscore = (mscore,)
            tmp_d, tmp_path = mkstemp(
                prefix=filename.stem, suffix=musicxml_c.MUSIC21_FILE_EXTENSIONS[0]
            )
            process = mscore + ("-fo", tmp_path, filename)
            res = subprocess.run(process, stdout=DEVNULL, stderr=DEVNULL)
            if res.returncode != 0:
                raise RuntimeError(
                    f"Error while converting musescore file to xml: {filename}"
                )
        else:
            tmp_path = filename
        score = parse_filename(
            tmp_path,
            self._cfg.split_keywords,
            expand_repeats=self._cfg.expand_repeats,
            export_dfs_to=self._cfg.dfs_dir,
            remove_unpitched_objects=self._cfg.remove_unpitched_objects,
        )
        numeric_tempo = extract_numeric_tempo(tmp_path)
        if filename.suffix == mscore_c.MUSESCORE_FILE_EXTENSION:
            os.close(tmp_d)
            os.remove(tmp_path)
        filtered_parts = self._filter_parts(score)
        return score, tuple(filtered_parts), numeric_tempo

    def _get_score_data(
        self, filename: PurePath, load_cache: Optional[Path] = None
    ) -> dict:
        data = None
        info_load_str = ""
        if load_cache is not None and load_cache.exists():
            try:
                data = pickle.load(open(load_cache, "rb"))
            except Exception as e:
                info_load_str += f" Error while loading pickled object, continuing with extraction from scratch: {e}"
            else:
                info_load_str += " File was loaded from cache."

        if data is None:
            try:
                score, filtered_parts, numeric_tempo = self._load_xml_data(filename)
            except ParseFileError as e:
                perr(f"Error while parsing file {filename}")
                raise e
            else:
                info_load_str += " XML file parsed succesfully!"
            if len(filtered_parts) == 0:
                lwarn(
                    f"No parts were found for file {filename} and filter: {','.join(self._cfg.parts_filter)}"
                )
            if (
                self._cfg.is_requested_musescore_file()
                and self._cfg.musescore_dir is not None
            ):
                filename_ms3 = (
                    Path(self._cfg.musescore_dir)
                    / filename.with_suffix(mscore_c.MUSESCORE_FILE_EXTENSION).name
                )
                try:
                    data_musescore = self._get_harmony_data(filename_ms3)
                except ParseFileError as e:
                    perr(f"Error while parsing file {filename_ms3}")
                    raise e
                else:
                    info_load_str += " MS3 file parsed succesfully!"
            else:
                data_musescore = None
            data = {
                C.DATA_SCORE: score,
                C.DATA_FILE: str(filename),
                C.DATA_FILTERED_PARTS: filtered_parts,
                C.DATA_MUSESCORE_SCORE: data_musescore,
                C.DATA_NUMERIC_TEMPO: numeric_tempo,
            }
            if len(self._cfg.precache_hooks) > 0:
                for hook in self._cfg.precache_hooks:
                    if isinstance(hook, str):
                        hook = __import__(hook, fromlist=[""])
                    hook.execute(self._cfg, data)
            if self._cfg.cache_dir is not None:
                m21_objects = SmartModuleCache(
                    (data[C.DATA_SCORE], data[C.DATA_FILTERED_PARTS]),
                    resurrect_reference=(
                        self._load_xml_data,
                        # filename.relative_to("."),
                        filename,
                    ),
                )
                data[C.DATA_SCORE] = m21_objects[0]
                data[C.DATA_FILTERED_PARTS] = m21_objects[1]

        pdebug(f"\nProcessing score {filename}." + info_load_str)
        return data

    def _get_harmony_data(self, filename: PurePath) -> pd.DataFrame:
        if not filename.exists():
            lerr(f"Musescore file was not found for {filename} file!")
            lerr(
                f"These features won't be extracted for {filename}: {C.REQUIRE_MSCORE}"
            )
            return None
        else:
            try:
                data_musescore = parse_musescore_file(
                    str(filename), self._cfg.expand_repeats
                )
                return data_musescore
            except ParseFileError as e:
                data_musescore = None
                lerr(str(e))
                return None

    def _filter_parts(self, score: Score) -> List[Part]:
        parts = list(score.parts)
        # self._deal_with_dupicated_parts(parts)
        if self._cfg.parts_filter is None or len(self._cfg.parts_filter) == 0:
            return parts
        filter_set = set(self._cfg.parts_filter)
        return (
            part
            for part in parts
            if to_abbreviation(part, parts, self._cfg) in filter_set
        )

    def _deal_with_dupicated_parts(self, parts):
        for part in parts:
            # Keeping onle solo and 1º part of duplicated instruments
            part.id = part.id.replace(" 1º", "")
            part.partAbbreviation = part.partAbbreviation.replace(" 1º", "")
            if "2º" in part.id:
                parts.remove(part)

    def _get_part_data(self, score_data: dict, part: Part) -> dict:
        sound = extract_sound(part, self._cfg)
        part_abbreviation, sound_abbreviation, part_number = _extract_abbreviated_part(
            sound, part, score_data[C.DATA_FILTERED_PARTS], self._cfg
        )
        family = self._cfg.sound_to_family.get(sound, GENERAL_FAMILY)
        family_abbreviation = self._cfg.family_to_abbreviation.get(family, family)
        data = {
            C.DATA_PART: part,
            C.DATA_PART_NUMBER: part_number,
            C.DATA_PART_ABBREVIATION: part_abbreviation,
            C.DATA_SOUND: sound,
            C.DATA_SOUND_ABBREVIATION: sound_abbreviation,
            C.DATA_FAMILY: family,
            C.DATA_FAMILY_ABBREVIATION: family_abbreviation,
        }
        return data

    def _get_module_or_attribute(self, f, name):
        if hasattr(f, name):
            module = getattr(f, name)
        else:
            try:
                module = __import__(f.__name__ + "." + name, fromlist=[""])
            except ModuleNotFoundError as e:
                return e
        return module

    def _find_modules(self, package: str, basic: bool):
        found_features = set()
        if isinstance(package, str):
            package = __import__(package, fromlist=[""])
        if basic:
            to_extract = self._cfg.basic_modules
        else:
            to_extract = self._cfg.features
        for feature in to_extract:
            feature_package = self._get_module_or_attribute(package, feature)
            if isinstance(feature_package, Exception):
                continue
            module = self._get_module_or_attribute(feature_package, "handler")
            if isinstance(module, Exception):
                raise ImportError(
                    f"It seems {feature}.handler cannot be imported."
                ) from module
            feature_dependencies = getattr(feature_package, "musif_dependencies", [])
            for dependency in feature_dependencies:
                if dependency not in found_features and dependency != feature:
                    raise ValueError(
                        f"Feature {feature} is dependent on feature {dependency} ({dependency} should appear before {feature} in the configuration)"
                    )
            found_features.add(feature)
            yield module

    def _update_parts_module_features(
        self,
        module,
        score_data: dict,
        parts_data: List[dict],
        parts_features: List[dict],
    ):
        for part_data, part_features in zip(parts_data, parts_features):
            module_name = (
                str(module.__name__)
                .replace("musif.extract.features.", "")
                .replace(".handler", "")
            )
            ldebug(
                f'Extracting part "{part_data[C.DATA_PART_ABBREVIATION]}" {module_name} features.'
            )
            try:
                module.update_part_objects(
                    score_data, part_data, self._cfg, part_features
                )
            except Exception as e:
                score_name = score_data["file"]
                perr(
                    f"An error occurred while extracting module {module.__name__} in {score_name}!!.\nError: {e}\n"
                )
                raise FeatureError(
                    f"In {score_name} while computing {module.__name__}"
                ) from e

    def _update_score_module_features(
        self,
        module,
        score_data: dict,
        parts_data: List[dict],
        parts_features: List[dict],
        score_features: dict,
    ):
        ldebug(
            f'Extracting score "{score_data[C.DATA_FILE]}" {module.__name__} features.'
        )
        try:
            module.update_score_objects(
                score_data, parts_data, self._cfg, parts_features, score_features
            )
        except Exception as e:
            score_name = score_data["file"]
            perr(
                f"An error occurred while extracting module {module.__name__} in {score_name}!!.\nError: {e}\n"
            )
            raise FeatureError(
                f"In {score_name} while computing {module.__name__}"
            ) from e



AVERAGE_DURATION = "AverageDuration"
RHYTHMINT = "RhythmInt"  # number of notes / number of beats
# RHYTHMINTSEP = "RhythmIntSep"  # SUM number of notes in a time signature/ beat in a time signature
DOTTEDRHYTHM = "DottedRhythm"
DOUBLE_DOTTEDRHYTHM = "DoubleDottedRhythm"




import logging
from logging import Logger, getLogger

from musif.common.constants import (
    LEVEL_CRITICAL,
    LEVEL_DEBUG,
    LEVEL_ERROR,
    LEVEL_INFO,
    LEVEL_WARNING,
)
from musif.common._utils import colorize
from musif.config import LOGGER_NAME

"""
This module contains functions that display information when using musif. 
Functionts starting with 'l' are using for logging, while function starting by
'p' are used for terminal display. The messages will be out puted acoording to the level set
in the configuration file.
"""


def linfo(text: str, exc_info: bool = False) -> None:
    llog(text, logger(), LEVEL_INFO, exc_info)


def ldebug(text: str, exc_info: bool = False) -> None:
    llog(text, logger(), LEVEL_DEBUG, exc_info)


def lwarn(text: str, exc_info: bool = False) -> None:
    llog(text, logger(), LEVEL_WARNING, exc_info)


def lerr(text: str, exc_info: bool = False) -> None:
    llog(text, logger(), LEVEL_ERROR, exc_info)


def lcrit(text: str, exc_info: bool = False) -> None:
    llog(text, logger(), LEVEL_CRITICAL, exc_info)


def pinfo(text: str, level: str = LEVEL_INFO) -> None:
    plog(text, LEVEL_INFO, allowed_level=level)


def pdebug(text: str, level: str = LEVEL_INFO) -> None:
    plog(text, LEVEL_DEBUG, allowed_level=level)


def pwarn(text: str, level: str = LEVEL_INFO) -> None:
    plog(text, LEVEL_WARNING, allowed_level=level)


def perr(text: str, level: str = LEVEL_INFO) -> None:
    plog(text, LEVEL_ERROR, allowed_level=level)


def pcrit(text: str, level: str = LEVEL_INFO) -> None:
    plog(text, LEVEL_CRITICAL, allowed_level=level)


def logger() -> Logger:
    return getLogger(LOGGER_NAME)


def llog(text: str, logger: logging.Logger, level: str, exc_info: bool) -> None:
    logger.log(logging.getLevelName(level), text, exc_info=exc_info)


def plog(text: str, level: str, allowed_level: str = LEVEL_INFO) -> None:
    if logging.getLevelName(level) < logging.getLevelName(allowed_level):
        return
    print(colorize(text, level))



from ntpath import basename
from typing import List

from musif.config import ExtractConfiguration
from musif.extract.constants import DATA_FILE
from musif.logs import lwarn
from .constants import *


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    """
    get variables from file_name
    returns a dictionary so it can be easily input in a df
    """

    file_name = basename(score_data[DATA_FILE])
    name = file_name.split('.')[:-1]
    if '_' in name:
        name = name.split('_')
        title = name[0]
        artist = ""
    else:
        title = "".join(name)
        artist = ""
    artist = name[0]
    title = name[1]
        
    score_features.update(
        {
            ARTIST: artist,
            TITLE: title,
        }
    )

def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    pass



from collections import Counter, defaultdict
from statistics import mean
from typing import List

import numpy as np

from musif.config import ExtractConfiguration
from musif.extract.constants import DATA_PART_ABBREVIATION, GLOBAL_TIME_SIGNATURE
from musif.extract.features.core.constants import DATA_NOTES
from musif.extract.features.prefix import get_part_feature, get_score_feature
from musif.extract.utils import _get_beat_position
from musif.musicxml.tempo import get_number_of_beats

from .constants import *


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):

    notes_duration = [note.duration.quarterLength for note in part_data["notes"]]
    beats = 1
    beat_unit = 1
    total_number_notes = 0
    number_notes = 0
    notes_dict = defaultdict(int)
    rhythm_intensity_period = []
    rhythm_dot = 0
    rhythm_double_dot = 0
    total_beats = 0
    total_sounding_beats = 0
    beat_count = (
        score_data[GLOBAL_TIME_SIGNATURE].beatCount
        if hasattr(score_data.get(GLOBAL_TIME_SIGNATURE), 'beatCount')
        else 1
    )
    # motion_features = get_motion_features(part_data)
    # part_features.update(motion_features)

    for measure in part_data["measures"]:
        for i, element in enumerate(measure.elements):
            if element.classes[0] == "Note":
                number_notes += 1
                notes_dict[element.duration.quarterLength] += 1
                total_number_notes += 1
                pos = _get_beat_position(beat_count, beats, element.beat)
                if (
                    element.duration.dots > 0
                    and element.duration.quarterLength < beat_unit
                ):
                    if i + 1 < len(measure.elements):  # check next item
                        if (
                            measure.elements[i + 1].beatStr.split()[0]
                            == element.beatStr.split()[0]
                        ):
                            if (
                                measure.elements[i + 1].duration.quarterLength
                                < element.duration.quarterLength
                            ):

                                rhythm_dot += 1

                    if element.duration.dots == 2:
                        rhythm_double_dot += 1
            elif element.classes[0] == "TimeSignature":
                rhythm_intensity_period.append(
                    sum([float(i) * j for i, j in Counter(notes_dict).items()])
                    / total_sounding_beats
                    if total_sounding_beats != 0
                    else 0
                )
                number_notes = 0
                beat_count = element.beatCount
                beats = get_number_of_beats(element.ratioString)
                beat_unit = element.beatStrength
        total_beats += beats

        if measure in part_data["sounding_measures"]:
            total_sounding_beats += beats

    rhythm_intensity_period.append(
        sum([j / i if i != 0.0 else 0 for i, j in Counter(notes_dict).items()])
        / total_sounding_beats
        if total_sounding_beats
        else 0
    )
    notes_duration = [
        i for i in notes_duration if i != 0.0
    ]  # remove notes with duration equal to 0

    part_features.update(
        {
            AVERAGE_DURATION: mean(notes_duration)
            if len(notes_duration) != 0
            else "NA",
            RHYTHMINT: sum(rhythm_intensity_period),
            DOTTEDRHYTHM: (rhythm_dot / total_sounding_beats)
            if total_sounding_beats
            else "NA",
            DOUBLE_DOTTEDRHYTHM: (rhythm_double_dot / total_beats),
        }
    )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    features = {}
    average_durations = []
    rhythm_intensities = []
    dotted_rhythm = []
    double_dotted_rhythm = []

    total_notes_duration = []

    for part_data, part_features in zip(parts_data, parts_features):
        part = part_data[DATA_PART_ABBREVIATION]

        features[get_part_feature(part, AVERAGE_DURATION)] = part_features[
            AVERAGE_DURATION
        ]
        total_notes_duration.append(
            [i.duration.quarterLength for i in part_data[DATA_NOTES] if i != 0]
        )

        features[get_part_feature(part, RHYTHMINT)] = part_features[RHYTHMINT]
        # rhythm_intensities.append([1/i.duration.quarterLength for i in part_data[DATA_NOTES] if i !=0])
        rhythm_intensities.append(part_features[RHYTHMINT])

        features[get_part_feature(part, DOTTEDRHYTHM)] = part_features[DOTTEDRHYTHM]
        dotted_rhythm.append(part_features[DOTTEDRHYTHM])

        features[get_part_feature(part, DOUBLE_DOTTEDRHYTHM)] = part_features[
            DOUBLE_DOTTEDRHYTHM
        ]
        double_dotted_rhythm.append(part_features[DOUBLE_DOTTEDRHYTHM])



    dotted_rhythm = [i for i in dotted_rhythm if i not in (0.0, "NA")]
    features.update(
        {
            # get_score_feature(AVERAGE_DURATION): mean(average_durations),
            get_score_feature(AVERAGE_DURATION): np.mean(
                [note for instrument in total_notes_duration for note in instrument]
            ),
            get_score_feature(RHYTHMINT): np.mean(rhythm_intensities)
            if rhythm_intensities
            else 0.0,
            # get_score_feature(RHYTHMINT): mean([rythm for instrument in rhythm_intensities for rythm in instrument])/score_features['Score_SoundingMeasures']*BEATS,
            get_score_feature(DOTTEDRHYTHM): np.mean(dotted_rhythm)
            if dotted_rhythm
            else 0.0,
            get_score_feature(DOUBLE_DOTTEDRHYTHM): np.mean(double_dotted_rhythm)
            if double_dotted_rhythm
            else 0.0,
        }
    )

    score_features.update(features)





import re
from statistics import mean

from typing import Dict, List, Union


from music21.note import Note


from musif.config import ExtractConfiguration

from musif.extract.common import _filter_parts_data

from musif.extract.features.core.handler import DATA_KEY, DATA_NOTES

from musif.extract.features.prefix import get_part_feature, get_score_feature

from musif.musicxml.common import _get_degrees_and_accidentals

from .constants import *

from ...constants import DATA_PART_ABBREVIATION


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):

    notes = part_data[DATA_NOTES]

    key = score_data[DATA_KEY]

    notes_per_degree = get_notes_per_degree(str(key), notes)

    all_degrees = sum(value for value in notes_per_degree.values())

    for key, value in notes_per_degree.items():

        part_features[DEGREE_COUNT.format(key=key)] = value

        part_features[DEGREE_PER.format(key=key)] = (
            value / all_degrees if all_degrees != 0 else 0
        )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    parts_data = _filter_parts_data(parts_data, cfg.parts_filter)

    if len(parts_data) == 0:
        return

    key = score_data[DATA_KEY]

    score_notes_per_degree = {}

    for part_data in parts_data:

        notes = part_data[DATA_NOTES]

        part_notes_per_degree = get_notes_per_degree(str(key), notes)

        for degree, notes in part_notes_per_degree.items():

            if degree not in score_notes_per_degree:

                score_notes_per_degree[degree] = 0

            score_notes_per_degree[degree] += notes

    all_score_degrees = sum(value for value in score_notes_per_degree.values())

    for key, value in score_notes_per_degree.items():

        score_features[get_score_feature(DEGREE_COUNT.format(key=key))] = value

        score_features[get_score_feature(DEGREE_PER.format(key=key))] = (
            value / all_score_degrees if all_score_degrees != 0 else 0
        )

    for part_data, parts_features in zip(parts_data, parts_features):

        part = part_data[DATA_PART_ABBREVIATION]

        degree_count_pattern = DEGREE_COUNT.format(key=".+")

        degree_per_pattern = DEGREE_PER.format(key=".+")

        for feature_name in parts_features:

            if re.match(degree_count_pattern, feature_name) or re.match(
                degree_per_pattern, feature_name
            ):

                part_feature = get_part_feature(part, feature_name)

                if part_feature in score_features:

                    score_features[part_feature] = mean(
                        [score_features[part_feature], parts_features[feature_name]]
                    )

                else:

                    score_features[part_feature] = parts_features[feature_name]


def get_notes_per_degree(key: str, notes: List[Note]) -> Dict[str, int]:

    notes_per_degree = {
        to_full_degree(degree, accidental): 0
        for accidental in ["", "sharp", "flat"]
        for degree in [1, 2, 3, 4, 5, 6, 7]
    }

    all_degrees = 0

    for degree, accidental in _get_degrees_and_accidentals(key, notes):

        if to_full_degree(degree, accidental) not in notes_per_degree:
            continue

        notes_per_degree[to_full_degree(degree, accidental)] += 1

        all_degrees += 1
    return notes_per_degree


def to_full_degree(degree: Union[int, str], accidental: str) -> str:

    return f"{ACCIDENTAL_ABBREVIATION[accidental]}{degree}"



from collections import Counter
from itertools import groupby
from statistics import mean, stdev
from typing import List, Tuple, Dict, Union

import numpy as np
import pandas as pd
from music21.interval import Interval
from scipy.stats import kurtosis, skew
from scipy.stats.mstats import trimmed_mean, trimmed_std

from musif.cache import hasattr
from musif.common._utils import extract_digits
from musif.config import ExtractConfiguration
from musif.extract.common import _mix_data_with_precedent_data
from musif.extract.constants import DATA_PART_ABBREVIATION, DATA_SOUND_ABBREVIATION
from musif.extract.features.core.constants import DATA_INTERVALS
from musif.extract.features.prefix import (
    get_part_feature,
    get_part_prefix,
    get_score_prefix,
    get_sound_prefix,
)

from .constants import *


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    intervals = part_data[DATA_INTERVALS]

    part_features.update(get_motion_features(part_data))
    part_features.update(get_interval_features(intervals))
    part_features.update(get_interval_count_features(intervals))
    part_features.update(get_interval_type_features(intervals))
    part_features.update(get_interval_stats_features(intervals))


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    if len(parts_data) == 0:
        return

    features = {}
    for part_data, part_features in zip(parts_data, parts_features):
        part = part_data[DATA_PART_ABBREVIATION]
        for step in MOTION_STEPS:
            for win in MOTION_WINS:
                if step > win:
                    continue
                key_postfix = _motion_postfix(step, win)
                features[
                    get_part_feature(part, SPEED_AVG_ABS + key_postfix)
                ] = part_features[SPEED_AVG_ABS + key_postfix]
                features[
                    get_part_feature(part, ACCELERATION_AVG_ABS + key_postfix)
                ] = part_features[ACCELERATION_AVG_ABS + key_postfix]
                features[
                    get_part_feature(part, ASCENDENT_AVERAGE + key_postfix)
                ] = part_features[ASCENDENT_AVERAGE + key_postfix]
                features[
                    get_part_feature(part, DESCENDENT_AVERAGE + key_postfix)
                ] = part_features[DESCENDENT_AVERAGE + key_postfix]
                features[
                    get_part_feature(part, ASCENDENT_PROPORTION + key_postfix)
                ] = part_features[ASCENDENT_PROPORTION + key_postfix]
                features[
                    get_part_feature(part, DESCENDENT_PROPORTION + key_postfix)
                ] = part_features[DESCENDENT_PROPORTION + key_postfix]

        part_prefix = get_part_prefix(part_data[DATA_PART_ABBREVIATION])
        intervals = part_data[DATA_INTERVALS]
        interval_features = get_interval_features(intervals, part_prefix)
        interval_count_features = get_interval_count_features(intervals, part_prefix)
        interval_type_features = get_interval_type_features(intervals, part_prefix)
        interval_stats_features = get_interval_stats_features(intervals, part_prefix)

        if all([i in features for i in interval_features.keys()]):
            _mix_data_with_precedent_data(features, interval_features)
        else:
            features.update(interval_features)
        if all([i in features for i in interval_count_features.keys()]):
            _mix_data_with_precedent_data(features, interval_count_features)
        else:
            features.update(interval_count_features)
        if all([i in features for i in interval_type_features.keys()]):
            _mix_data_with_precedent_data(features, interval_type_features)
        else:
            features.update(interval_type_features)
        if all([i in features for i in interval_stats_features.keys()]):
            _mix_data_with_precedent_data(features, interval_stats_features)
        else:
            features.update(interval_stats_features)

    parts_data_per_sound = {
        part_data[DATA_SOUND_ABBREVIATION]: [] for part_data in parts_data
    }
    for part_data in parts_data:
        sound = part_data[DATA_SOUND_ABBREVIATION]
        parts_data_per_sound[sound].append(part_data)

    for sound, sound_parts_data in parts_data_per_sound.items():
        sound_prefix = get_sound_prefix(sound)
        intervals = [
            interval
            for part_data in sound_parts_data
            for interval in part_data[DATA_INTERVALS]
        ]
        features.update(get_interval_features(intervals, sound_prefix))
        features.update(get_interval_count_features(intervals, sound_prefix))
        features.update(get_interval_type_features(intervals, sound_prefix))
        features.update(get_interval_stats_features(intervals, sound_prefix))

    score_intervals = [
        interval for part_data in parts_data for interval in part_data[DATA_INTERVALS]
    ]
    score_prefix = get_score_prefix()

    features.update(get_interval_features(score_intervals, score_prefix))
    features.update(get_interval_count_features(score_intervals, score_prefix))
    features.update(get_interval_type_features(score_intervals, score_prefix))
    features.update(get_interval_stats_features(score_intervals, score_prefix))

    score_features.update(features)


def get_interval_features(intervals: List[Interval], prefix: str = ""):
    numeric_intervals = [interval.semitones for interval in intervals]
    absolute_numeric_intervals = [
        abs(numeric_interval) for numeric_interval in numeric_intervals
    ]
    ascending_intervals = [
        numeric_interval
        for numeric_interval in numeric_intervals
        if numeric_interval > 0
    ]
    descending_intervals = [
        numeric_interval
        for numeric_interval in numeric_intervals
        if numeric_interval < 0
    ]

    absolute_intervallic_mean = (
        mean(absolute_numeric_intervals) if len(intervals) > 0 else 0
    )
    intervallic_mean = mean(numeric_intervals) if len(intervals) > 0 else 0
    ascending_intervallic_mean = (
        mean(ascending_intervals) if len(ascending_intervals) > 0 else 0
    )
    descending_intervallic_mean = (
        mean(descending_intervals) if len(descending_intervals) > 0 else 0
    )
    with np.errstate(invalid="ignore"):
        absolute_intervallic_std = (
            stdev(absolute_numeric_intervals) if len(intervals) > 1 else 0
        )
        intervallic_std = stdev(numeric_intervals) if len(intervals) > 1 else 0
        ascending_intervallic_std = (
            stdev(ascending_intervals) if len(ascending_intervals) > 1 else 0
        )
        descending_intervallic_std = (
            stdev(descending_intervals) if len(descending_intervals) > 1 else 0
        )
    mean_interval = Interval(int(round(absolute_intervallic_mean))).directedName

    cutoff = 0.1
    limits = (cutoff, cutoff)
    trimmed_intervallic_mean = (
        trimmed_mean(numeric_intervals, limits) if len(intervals) > 0 else 0
    )
    trimmed_absolute_intervallic_mean = (
        trimmed_mean(absolute_numeric_intervals, limits) if len(intervals) > 0 else 0
    )
    with np.errstate(invalid="ignore"):
        trimmed_intervallic_std = (
            trimmed_std(numeric_intervals, limits) if len(intervals) > 1 else 0
        )
        trimmed_absolute_intervallic_std = (
            trimmed_std(absolute_numeric_intervals, limits) if len(intervals) > 1 else 0
        )
    trim_diff = intervallic_mean - trimmed_intervallic_mean
    trim_ratio = trim_diff / intervallic_mean if intervallic_mean != 0 else 0
    absolute_trim_diff = absolute_intervallic_mean - trimmed_absolute_intervallic_mean
    absolute_trim_ratio = (
        absolute_trim_diff / absolute_intervallic_mean
        if absolute_intervallic_mean != 0
        else 0
    )
    num_ascending_intervals = (
        len(ascending_intervals) if len(ascending_intervals) > 0 else 0
    )
    num_descending_intervals = (
        len(descending_intervals) if len(descending_intervals) > 0 else 0
    )
    num_ascending_semitones = (
        sum(ascending_intervals) if len(ascending_intervals) > 0 else 0
    )
    num_descending_semitones = (
        sum(descending_intervals) if len(descending_intervals) > 0 else 0
    )
    ascending_intervals_percentage = (
        num_ascending_intervals / len(intervals) if len(intervals) > 0 else 0
    )
    descending_intervals_percentage = (
        num_descending_intervals / len(intervals) if len(intervals) > 0 else 0
    )

    largest_semitones = max(numeric_intervals) if len(numeric_intervals) > 0 else None
    largest = (
        Interval(largest_semitones).directedName if len(numeric_intervals) > 0 else None
    )
    smallest_semitones = (
        sorted(numeric_intervals, key=abs)[0] if len(numeric_intervals) > 0 else None
    )
    smallest = Interval(smallest_semitones).directedName if len(intervals) > 0 else None
    largest_ascending_semitones = (
        max(ascending_intervals) if len(ascending_intervals) > 0 else None
    )
    largest_ascending = (
        Interval(largest_ascending_semitones).directedName
        if len(ascending_intervals) > 0
        else None
    )
    largest_descending_semitones = (
        min(descending_intervals) if len(descending_intervals) > 0 else None
    )
    largest_descending = (
        Interval(largest_descending_semitones).directedName
        if len(descending_intervals) > 0
        else None
    )
    smallest_ascending_semitones = (
        min(ascending_intervals) if len(ascending_intervals) > 0 else None
    )
    smallest_ascending = (
        Interval(smallest_ascending_semitones).directedName
        if len(ascending_intervals) > 0
        else None
    )
    smallest_descending_semitones = (
        max(descending_intervals) if len(descending_intervals) > 0 else None
    )
    smallest_descending = (
        Interval(smallest_descending_semitones).directedName
        if len(descending_intervals) > 0
        else None
    )

    features = {
        f"{prefix}{MEAN_INTERVAL}": mean_interval,
        f"{prefix}{INTERVALLIC_MEAN}": intervallic_mean,
        f"{prefix}{INTERVALLIC_STD}": intervallic_std,
        f"{prefix}{ABSOLUTE_INTERVALLIC_MEAN}": absolute_intervallic_mean,
        f"{prefix}{ABSOLUTE_INTERVALLIC_STD}": absolute_intervallic_std,
        f"{prefix}{ASCENDING_INTERVALLIC_MEAN}": ascending_intervallic_mean,
        f"{prefix}{ASCENDING_INTERVALLIC_STD}": ascending_intervallic_std,
        f"{prefix}{DESCENDING_INTERVALLIC_MEAN}": descending_intervallic_mean,
        f"{prefix}{DESCENDING_INTERVALLIC_STD}": descending_intervallic_std,
        f"{prefix}{TRIMMED_INTERVALLIC_MEAN}": trimmed_intervallic_mean,
        f"{prefix}{TRIMMED_INTERVALLIC_STD}": trimmed_intervallic_std,
        f"{prefix}{TRIMMED_ABSOLUTE_INTERVALLIC_MEAN}": trimmed_absolute_intervallic_mean,
        f"{prefix}{TRIMMED_ABSOLUTE_INTERVALLIC_STD}": trimmed_absolute_intervallic_std,
        f"{prefix}{INTERVALLIC_TRIM_DIFF}": trim_diff,
        f"{prefix}{INTERVALLIC_TRIM_RATIO}": trim_ratio,
        f"{prefix}{ABSOLUTE_INTERVALLIC_TRIM_DIFF}": absolute_trim_diff,
        f"{prefix}{ABSOLUTE_INTERVALLIC_TRIM_RATIO}": absolute_trim_ratio,
        f"{prefix}{ASCENDING_INTERVALS_COUNT}": num_ascending_intervals,
        f"{prefix}{DESCENDING_INTERVALS_COUNT}": num_descending_intervals,
        f"{prefix}{ASCENDING_SEMITONES_SUM}": num_ascending_semitones,
        f"{prefix}{DESCENDING_SEMITONES_SUM}": num_descending_semitones,
        f"{prefix}{ASCENDING_INTERVALS_PER}": ascending_intervals_percentage,
        f"{prefix}{DESCENDING_INTERVALS_PER}": descending_intervals_percentage,
        f"{prefix}{LARGEST_INTERVAL_ALL}": largest,
        f"{prefix}{LARGEST_INTERVAL_ASC}": largest_ascending,
        f"{prefix}{LARGEST_INTERVAL_DESC}": largest_descending,
        f"{prefix}{LARGEST_SEMITONES_ALL}": largest_semitones,
        f"{prefix}{LARGEST_SEMITONES_ASC}": largest_ascending_semitones,
        f"{prefix}{LARGEST_SEMITONES_DESC}": largest_descending_semitones,
        f"{prefix}{LARGEST_ABSOLUTE_SEMITONES_ALL}": abs(largest_semitones)
        if largest_semitones is not None
        else None,
        f"{prefix}{LARGEST_ABSOLUTE_SEMITONES_ASC}": abs(largest_ascending_semitones)
        if largest_ascending_semitones is not None
        else None,
        f"{prefix}{LARGEST_ABSOLUTE_SEMITONES_DESC}": abs(largest_descending_semitones)
        if largest_descending_semitones is not None
        else None,
        f"{prefix}{SMALLEST_INTERVAL_ALL}": smallest,
        f"{prefix}{SMALLEST_INTERVAL_ASC}": smallest_ascending,
        f"{prefix}{SMALLEST_INTERVAL_DESC}": smallest_descending,
        f"{prefix}{SMALLEST_SEMITONES_ALL}": smallest_semitones,
        f"{prefix}{SMALLEST_SEMITONES_ASC}": smallest_ascending_semitones,
        f"{prefix}{SMALLEST_SEMITONES_DESC}": smallest_descending_semitones,
        f"{prefix}{SMALLEST_ABSOLUTE_SEMITONES_ALL}": abs(smallest_semitones)
        if smallest_semitones is not None
        else None,
        f"{prefix}{SMALLEST_ABSOLUTE_SEMITONES_ASC}": abs(smallest_ascending_semitones)
        if smallest_ascending_semitones is not None
        else None,
        f"{prefix}{SMALLEST_ABSOLUTE_SEMITONES_DESC}": abs(
            smallest_descending_semitones
        )
        if smallest_descending_semitones is not None
        else None,
    }
    return features


def get_interval_count_features(intervals: List[Interval], prefix: str = "") -> dict:
    interval_counts = Counter([interval.directedName for interval in intervals])
    total_count = len(intervals)
    interval_features = {}
    for interval, count in interval_counts.items():
        interval_features[
            INTERVAL_COUNT.format(prefix=prefix, interval=interval)
        ] = count
        interval_features[INTERVAL_PER.format(prefix=prefix, interval=interval)] = (
            count / total_count if total_count else 0
        )
    return interval_features


def get_interval_type_features(intervals_list: List[Interval], prefix: str = ""):
    repeated_notes_list = []
    stepwise_list = []
    leaps_list = []
    within_octave_list = []
    beyond_octave_list = []
    perfect_list = []
    major_list = []
    minor_list = []
    double_augmented_list = []
    augmented_list = []
    double_diminished_list = []
    diminished_list = []
    for interval in intervals_list:
        name = interval.directedName
        interval_number = int(extract_digits(name))
        if interval_number == 1:
            repeated_notes_list.append(interval)
        elif interval_number == 2:
            stepwise_list.append(interval)
        elif interval_number >= 3:
            leaps_list.append(interval)

        if abs(interval.semitones) <= 12:
            within_octave_list.append(interval)
        else:
            beyond_octave_list.append(interval)

        if name.startswith("AA"):
            double_augmented_list.append(interval)
        elif name.startswith("A"):
            augmented_list.append(interval)
        elif name.startswith("M"):
            major_list.append(interval)
        elif name.lower().startswith("p"):
            perfect_list.append(interval)
        elif name.startswith("m"):
            minor_list.append(interval)
        elif name.startswith("dd"):
            double_diminished_list.append(interval)
        elif name.startswith("d"):
            diminished_list.append(interval)
        else:
            raise ValueError(f"Unexpected interval name: {name}")
    all_intervals = len(intervals_list)
    all_repeated = len(repeated_notes_list)
    all_stepwise, ascending_stepwise, descending_stepwise = get_all_asc_desc_count(
        stepwise_list
    )
    all_leaps, ascending_leaps, descending_leaps = get_all_asc_desc_count(leaps_list)
    (
        all_within_octave,
        ascending_within_octave,
        descending_within_octave,
    ) = get_all_asc_desc_count(within_octave_list)
    (
        all_beyond_octave,
        ascending_beyond_octave,
        descending_beyond_octave,
    ) = get_all_asc_desc_count(beyond_octave_list)
    (
        all_double_augmented,
        ascending_double_augmented,
        descending_double_augmented,
    ) = get_all_asc_desc_count(double_augmented_list)
    all_augmented, ascending_augmented, descending_augmented = get_all_asc_desc_count(
        augmented_list
    )
    all_major, ascending_major, descending_major = get_all_asc_desc_count(major_list)
    all_perfect, ascending_perfect, descending_perfect = get_all_asc_desc_count(
        perfect_list
    )
    all_minor, ascending_minor, descending_minor = get_all_asc_desc_count(minor_list)
    (
        all_diminished,
        ascending_diminished,
        descending_diminished,
    ) = get_all_asc_desc_count(diminished_list)
    (
        all_double_diminished,
        ascending_double_diminished,
        descending_double_diminished,
    ) = get_all_asc_desc_count(double_diminished_list)

    return {
        f"{prefix}{REPEATED_NOTES_COUNT}": all_repeated,
        f"{prefix}{REPEATED_NOTES_PER}": all_repeated / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{STEPWISE_MOTION_ASC_COUNT}": ascending_stepwise,
        f"{prefix}{STEPWISE_MOTION_DESC_COUNT}": descending_stepwise,
        f"{prefix}{STEPWISE_MOTION_ALL_COUNT}": all_stepwise,
        f"{prefix}{STEPWISE_MOTION_ASC_PER}": ascending_stepwise / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{STEPWISE_MOTION_DESC_PER}": descending_stepwise / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{STEPWISE_MOTION_ALL_PER}": all_stepwise / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{LEAPS_ASC_COUNT}": ascending_leaps,
        f"{prefix}{LEAPS_DESC_COUNT}": descending_leaps,
        f"{prefix}{LEAPS_ALL_COUNT}": all_leaps,
        f"{prefix}{LEAPS_ASC_PER}": ascending_leaps / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{LEAPS_DESC_PER}": descending_leaps / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{LEAPS_ALL_PER}": all_leaps / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_PERFECT_ASC_COUNT}": ascending_perfect,
        f"{prefix}{INTERVALS_PERFECT_DESC_COUNT}": descending_perfect,
        f"{prefix}{INTERVALS_PERFECT_ALL_COUNT}": all_perfect,
        f"{prefix}{INTERVALS_PERFECT_ASC_PER}": ascending_perfect / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_PERFECT_DESC_PER}": descending_perfect / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_PERFECT_ALL_PER}": all_perfect / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_MAJOR_ASC_COUNT}": ascending_major,
        f"{prefix}{INTERVALS_MAJOR_DESC_COUNT}": descending_major,
        f"{prefix}{INTERVALS_MAJOR_ALL_COUNT}": all_major,
        f"{prefix}{INTERVALS_MAJOR_ASC_PER}": ascending_major / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_MAJOR_DESC_PER}": descending_major / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_MAJOR_ALL_PER}": all_major / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_MINOR_ASC_COUNT}": ascending_minor,
        f"{prefix}{INTERVALS_MINOR_DESC_COUNT}": descending_minor,
        f"{prefix}{INTERVALS_MINOR_ALL_COUNT}": all_minor,
        f"{prefix}{INTERVALS_MINOR_ASC_PER}": ascending_minor / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_MINOR_DESC_PER}": descending_minor / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_MINOR_ALL_PER}": all_minor / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_AUGMENTED_ASC_COUNT}": ascending_augmented,
        f"{prefix}{INTERVALS_AUGMENTED_DESC_COUNT}": descending_augmented,
        f"{prefix}{INTERVALS_AUGMENTED_ALL_COUNT}": all_augmented,
        f"{prefix}{INTERVALS_AUGMENTED_ASC_PER}": ascending_augmented / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_AUGMENTED_DESC_PER}": descending_augmented / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_AUGMENTED_ALL_PER}": all_augmented / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DIMINISHED_ASC_COUNT}": ascending_diminished,
        f"{prefix}{INTERVALS_DIMINISHED_DESC_COUNT}": descending_diminished,
        f"{prefix}{INTERVALS_DIMINISHED_ALL_COUNT}": all_diminished,
        f"{prefix}{INTERVALS_DIMINISHED_ASC_PER}": ascending_diminished / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DIMINISHED_DESC_PER}": descending_diminished
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DIMINISHED_ALL_PER}": all_diminished / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DOUBLE_AUGMENTED_ASC_COUNT}": ascending_double_augmented,
        f"{prefix}{INTERVALS_DOUBLE_AUGMENTED_DESC_COUNT}": descending_double_augmented,
        f"{prefix}{INTERVALS_DOUBLE_AUGMENTED_ALL_COUNT}": all_double_augmented,
        f"{prefix}{INTERVALS_DOUBLE_AUGMENTED_ASC_PER}": ascending_double_augmented
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DOUBLE_AUGMENTED_DESC_PER}": descending_double_augmented
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DOUBLE_AUGMENTED_ALL_PER}": all_double_augmented
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DOUBLE_DIMINISHED_ASC_COUNT}": ascending_double_diminished,
        f"{prefix}{INTERVALS_DOUBLE_DIMINISHED_DESC_COUNT}": descending_double_diminished,
        f"{prefix}{INTERVALS_DOUBLE_DIMINISHED_ALL_COUNT}": all_double_diminished,
        f"{prefix}{INTERVALS_DOUBLE_DIMINISHED_ASC_PER}": ascending_double_diminished
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DOUBLE_DIMINISHED_DESC_PER}": descending_double_diminished
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_DOUBLE_DIMINISHED_ALL_PER}": all_double_diminished
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_WITHIN_OCTAVE_ASC_COUNT}": ascending_within_octave,
        f"{prefix}{INTERVALS_WITHIN_OCTAVE_DESC_COUNT}": descending_within_octave,
        f"{prefix}{INTERVALS_WITHIN_OCTAVE_ALL_COUNT}": all_within_octave,
        f"{prefix}{INTERVALS_WITHIN_OCTAVE_ASC_PER}": ascending_within_octave
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_WITHIN_OCTAVE_DESC_PER}": descending_within_octave
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_WITHIN_OCTAVE_ALL_PER}": all_within_octave / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_BEYOND_OCTAVE_ASC_COUNT}": ascending_beyond_octave,
        f"{prefix}{INTERVALS_BEYOND_OCTAVE_DESC_COUNT}": descending_beyond_octave,
        f"{prefix}{INTERVALS_BEYOND_OCTAVE_ALL_COUNT}": all_beyond_octave,
        f"{prefix}{INTERVALS_BEYOND_OCTAVE_ASC_PER}": ascending_beyond_octave
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_BEYOND_OCTAVE_DESC_PER}": descending_beyond_octave
        / all_intervals
        if all_intervals != 0
        else 0,
        f"{prefix}{INTERVALS_BEYOND_OCTAVE_ALL_PER}": all_beyond_octave / all_intervals
        if all_intervals != 0
        else 0,
    }


def get_ascending_descending(
    intervals: List[Interval],
) -> Tuple[List[Interval], List[Interval]]:
    ascending = [interval for interval in intervals if interval.semitones > 0]
    descending = [interval for interval in intervals if interval.semitones < 0]
    return ascending, descending


def get_all_asc_desc_count(intervals: List[Interval]) -> Tuple[int, int, int]:
    ascending, descending = get_ascending_descending(intervals)
    return len(intervals), len(ascending), len(descending)


def get_interval_stats_features(intervals: List[Interval], prefix: str = ""):
    numeric_intervals = np.array([interval.semitones for interval in intervals])
    absolute_numeric_intervals = abs(numeric_intervals)
    with np.errstate(invalid="ignore"):
        intervals_skewness = (
            skew(numeric_intervals, bias=False)
            if [i for i in numeric_intervals if i != 0]
            else None
        )
        intervals_kurtosis = (
            kurtosis(numeric_intervals, bias=False)
            if [i for i in numeric_intervals if i != 0]
            else None
        )
        absolute_intervals_skewness = (
            skew(absolute_numeric_intervals, bias=False)
            if [i for i in absolute_numeric_intervals if i != 0]
            else None
        )
        absolute_intervals_kurtosis = (
            kurtosis(absolute_numeric_intervals, bias=False)
            if [i for i in absolute_numeric_intervals if i != 0]
            else None
        )

    return {
        f"{prefix}{INTERVALLIC_SKEWNESS}": intervals_skewness,
        f"{prefix}{INTERVALLIC_KURTOSIS}": intervals_kurtosis,
        f"{prefix}{ABSOLUTE_INTERVALLIC_SKEWNESS}": absolute_intervals_skewness,
        f"{prefix}{ABSOLUTE_INTERVALLIC_KURTOSIS}": absolute_intervals_kurtosis,
    }


def _motion_postfix(step, win):
    key_postfix = f"_step_{step}_win_{win}"
    return key_postfix


def _motion_features_single_window_step(
    notes_duration: List[float], notes_midi: List[int], step: float, win: int
) -> Dict[str, Union[float, np.ndarray]]:
    """
    Calculates motion features for a single window step of an aria.

    Parameters:
    notes_duration (List[float]): List of note durations in seconds.
    notes_midi (List[int]): List of MIDI note numbers.
    step (float): Step size in seconds.
    win (int): Window size in compasses.

    Returns:
    Dict[str, Union[float, np.ndarray]]: Dictionary containing the following motion features:
        - SPEED_AVG_ABS_{step}_{win} Average absolute speed.
        - ACCELERATION_AVG_ABS_{step}_{win} Average absolute acceleration.
        - ASCENDENT_AVERAGE_{step}_{win} Average length of prolonged ascent chunks in the smoothed midis of the aria.
        - DESCENDENT_AVERAGE_{step}_{win} Average length of prolonged descent chunks in the smoothed midis of the aria.
        - ASCENDENT_PROPORTION_{step}_{win} Proportion of prolonged ascent chunks over the total of the aria.
        - DESCENDENT_PROPORTION_{step}_{win} Proportion of prolonged descent chunks over the total of the aria.
    """
    key_postfix = _motion_postfix(step, win)
    default_dict = {
        SPEED_AVG_ABS + key_postfix: 0,
        ACCELERATION_AVG_ABS + key_postfix: 0,
        ASCENDENT_AVERAGE + key_postfix: 0,
        DESCENDENT_AVERAGE + key_postfix: 0,
        ASCENDENT_PROPORTION + key_postfix: 0,
        DESCENDENT_PROPORTION + key_postfix: 0,
    }
    
    if len(notes_midi) == 0:
        return default_dict
    midis_raw = np.repeat(notes_midi, np.divide(notes_duration, step).astype(int), axis=0)
    if midis_raw.size == 0:
        return default_dict

    # Absolute means of speed and acceleration
    spe_raw = np.diff(midis_raw) / step
    if spe_raw.size > 0:
        spe_avg_abs = np.mean(abs(spe_raw))
    else:
        spe_avg_abs = 0
    acc_raw = np.diff(spe_raw) / step
    if acc_raw.size > 0:
        acc_avg_abs = np.mean(abs(acc_raw))
    else:
        acc_avg_abs = 0

    # Rolling mean to smooth the midis by +-1 compasses -- not required for
    # statistics based on means but important for detecting increasing sequences
    # with a tolerance.
    midis_smo_series = pd.Series(midis_raw)
    midis_smo = [
        np.mean(i.to_list()) for i in midis_smo_series.rolling(2 * win + 1, center=True)
    ]

    # Prolonged ascent/descent chunks in smoothed midis of the aria (allows for
    # small violations in the form of decrements/increments that do not
    # decrease/increase the rolling mean).
    dife = np.diff(midis_smo)

    asc = [(k, sum(1 for i in g)) for k, g in groupby(dife > 0)]
    dsc = [(k, sum(1 for i in g)) for k, g in groupby(dife < 0)]

    asc = [i for b, i in asc if b]
    dsc = [i for b, i in dsc if b]

    # Average length of ascent/descent chunks of the aria
    asc_avg = mean(asc) if asc else np.nan
    dsc_avg = mean(dsc) if dsc else np.nan

    # Proportion of ascent/descent chunks over the total of the aria
    asc_prp = sum(asc) / (len(dife) - 1) if asc else np.nan
    dsc_prp = sum(dsc) / (len(dife) - 1) if dsc else np.nan

    return {
        SPEED_AVG_ABS + key_postfix: spe_avg_abs,
        ACCELERATION_AVG_ABS + key_postfix: acc_avg_abs,
        ASCENDENT_AVERAGE + key_postfix: asc_avg,
        DESCENDENT_AVERAGE + key_postfix: dsc_avg,
        ASCENDENT_PROPORTION + key_postfix: asc_prp,
        DESCENDENT_PROPORTION + key_postfix: dsc_prp,
    }


def get_motion_features(part_data) -> dict:
    """
    Extracts motion features from the given part data.

    Parameters:
    part_data (dict): A dictionary containing the notes and rests of a music part.

    Returns:
    dict: A dictionary containing the extracted motion features.

    Raises:
    This function does not raise any exceptions.
    """
    notes_midi = []
    notes_duration = []
    for note in part_data["notes_and_rests"]:
        if hasattr(note, "pitch"):
            notes_midi.append(note.pitch.midi)
            notes_duration.append(note.duration.quarterLength)

    notes_midi = np.asarray(notes_midi)
    notes_duration = np.asarray(notes_duration)

    return_dict = {}
    for step in MOTION_STEPS:
        for win in MOTION_WINS:
            return_dict.update(
                _motion_features_single_window_step(
                    notes_duration, notes_midi, step, win
                )
            )
    return return_dict



from ms3.score import MSCX

from musif.logs import ldebug, lerr


def extract_harmonic_analysis(mscx_file: str):
    """
    Given a mscx file name, parses the file using ms3 library and returns a dataframe containing all harmonic information.
    Adds Playthrough column that contains number of every measure in the cronological order
    Parameters
    ----------
    file_path: str
        Path to mscx file
    expand_repeats: bool
        Directory path to musescore file
    Returns
    -------
    harmonic_analysis: str
        Dataframe containing harmonic information

    """
    ldebug(f"Extracting harmonic analysis from musescore file '{mscx_file}'")
    harmonic_analysis = None
    try:
        musescore_score = MSCX(mscx_file, level="c")
        harmonic_analysis = musescore_score.expanded
    except Exception as e:
        lerr(f"Error found while parsing the musescore file {mscx_file} {str(e)}")
    return harmonic_analysis



ACCIDENTAL_ABBREVIATION = {
    "": "",
    "sharp": "#",
    "flat": "b",
    "double-sharp": "x",
    "double-flat": "bb",
    "triple-flat": "bbb",
    "triple-sharp": "#x"
}
DEGREE_PREFIX = "Degree"
DEGREE_COUNT = "Degree{key}_Count"
DEGREE_PER = "Degree{key}_Per"



from musif.common._constants import ENCODING
from musif.extract import FeaturesExtractor
from musif.process.processor import DataProcessor
import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except:
    pass



DATA_PART = "part"
DATA_PART_NUMBER = "part_number"
DATA_PART_ABBREVIATION = "abbreviation"
DATA_SOUND = "sound"
DATA_SOUND_ABBREVIATION = "sound_abbreviation"
DATA_FAMILY = "family"
DATA_FAMILY_ABBREVIATION = "family_abbreviation"
DATA_SCORE = "score"
DATA_MUSESCORE_SCORE = "MS3_score"
DATA_FILE = "file"
DATA_FILTERED_PARTS = "parts"
DATA_NUMERIC_TEMPO = "numeric_tempo"

HARMONY_FEATURES = "harmony"
SCALE_RELATIVE_FEATURES = "scale_relative"
REQUIRE_MSCORE = [HARMONY_FEATURES, SCALE_RELATIVE_FEATURES]
"""Names of modules taht require harmonic analysis in a .mscx file"""

VOICES_LIST = ["sop", "ten", "alt", "bar", "bbar", "bass"]
"""List of prefixes of singers's names that might appear in the scores"""

PLAYTHROUGH = "playthrough"
"""Constant for playthrough (count fo measures) added to ms3 dataframe"""

GLOBAL_TIME_SIGNATURE = "global_ts"
"""The name used for the column indicating the global time signature"""

WINDOW_RANGE = "WindowRange"
"""The name used for the column indicating the start and end of a window"""

WINDOW_ID = "WindowId"
"""The name used for the column of the window's id"""

ID = "Id"
"""The name used for the column of the music score's id"""



from typing import List

from pandas import DataFrame

from musif.config import ExtractConfiguration
from musif.extract.features.core.handler import DATA_MODE
from musif.logs import perr
from .constants import *
from .utils import (
    get_additions,
    get_chord_types,
    get_chords,
    get_harmonic_rhythm,
    get_keyareas,
    get_numerals,
)
from ...constants import DATA_MUSESCORE_SCORE


def get_harmony_data(score_features: dict, harmonic_analysis: DataFrame) -> dict:

    harmonic_rhythm = get_harmonic_rhythm(harmonic_analysis)
    numerals = get_numerals(harmonic_analysis)
    chord_types = get_chord_types(harmonic_analysis)
    additions = get_additions(harmonic_analysis)
    return dict(**harmonic_rhythm, **numerals, **chord_types, **additions)


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    features = {}
    try:
        harmonic_analysis = score_data.get(DATA_MUSESCORE_SCORE)

        if harmonic_analysis is None:
            features[HARMONY_AVAILABLE] = 0
            return features
        else:
            features[HARMONY_AVAILABLE] = 1

        all_harmonic_info = get_harmony_data(score_features, harmonic_analysis)
        keyareas = get_keyareas(
            harmonic_analysis, major=score_data[DATA_MODE] == "major"
        )
        chords, chords_grouping1, chords_grouping2 = get_chords(harmonic_analysis)

        features[f"{HARMONIC_RHYTHM}"] = all_harmonic_info[HARMONIC_RHYTHM]
        features[f"{HARMONIC_RHYTHM_BEATS}"] = all_harmonic_info[HARMONIC_RHYTHM_BEATS]
        features.update(
            {
                k: v
                for (k, v) in all_harmonic_info.items()
                if k.startswith(NUMERALS_prefix)
            }
        )
        features.update({k: v for (k, v) in keyareas.items()})
        features.update(
            {
                k: v
                for (k, v) in all_harmonic_info.items()
                if k.startswith(CHORD_TYPES_prefix)
            }
        )
        features.update(
            {k: v for (k, v) in chords.items() if k.startswith(CHORD_prefix)}
        )
        features.update({k: v for (k, v) in chords_grouping1.items()})
        features.update({k: v for (k, v) in chords_grouping2.items()})

        features.update(
            {
                k: v
                for (k, v) in all_harmonic_info.items()
                if k.startswith(ADDITIONS_prefix)
            }
        )

    except Exception as e:
        perr(f"Harmony problem found: {str(e)}")

    finally:
        score_features.update(features)


def update_part_objects(score_data, part_data, cfg, part_features):
    pass



from typing import List

from musif.common._constants import GENERAL_FAMILY, VOICE_FAMILY
from musif.common.sort import sort_list
from musif.config import ExtractConfiguration
from musif.extract.common import _part_matches_filter
from musif.extract.constants import (
    DATA_FAMILY,
    DATA_FAMILY_ABBREVIATION,
    DATA_FILTERED_PARTS,
    DATA_PART,
    DATA_PART_ABBREVIATION,
    DATA_PART_NUMBER,
    DATA_SCORE,
    DATA_SOUND,
    DATA_SOUND_ABBREVIATION,
)
from musif.extract.features.prefix import (
    get_family_feature,
    get_family_prefix,
    get_score_feature,
    get_sound_feature,
    get_sound_prefix,
)
from musif.musicxml.scoring import (
    ROMAN_NUMERALS_FROM_1_TO_20,
    _extract_abbreviated_part,
    extract_sound,
)
from .constants import *


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    part = part_data[DATA_PART]
    instrumental = part_data[DATA_FAMILY] != VOICE_FAMILY
    part_features.update(
        {
            PART_NAME: part.partName,
            PART_NUMBER: part_data[DATA_PART_NUMBER],
            PART_ABBREVIATION: part_data[DATA_PART_ABBREVIATION],
            SOUND: part_data[DATA_SOUND],
            SOUND_ABBREVIATION: part_data[DATA_SOUND_ABBREVIATION],
            FAMILY: part_data[DATA_FAMILY],
            FAMILY_ABBREVIATION: part_data[DATA_FAMILY_ABBREVIATION],
            INSTRUMENTAL: instrumental,
        }
    )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    abbreviated_parts = []
    sounds = []
    sound_abbreviations = []
    instrument_abbreviations = []
    voice_abbreviations = []
    families = []
    family_abbreviations = []
    instrumental_family_abbreviations = []
    count_by_sound = {}
    count_by_family = {}
    filtered_count_by_sound = {}
    filtered_count_by_family = {}
    score = score_data[DATA_SCORE]
    for part in score.parts:
        sound = extract_sound(part, cfg)
        part_abbreviation, sound_abbreviation, part_number = _extract_abbreviated_part(
            sound, part, score_data[DATA_FILTERED_PARTS], cfg)
        is_matching_part = _part_matches_filter(part_abbreviation, cfg.parts_filter)
        family = cfg.sound_to_family.get(sound, GENERAL_FAMILY)
        family_abbreviation = cfg.family_to_abbreviation.get(family, family)
        abbreviated_parts.append(part_abbreviation)
        instrumental = family != VOICE_FAMILY
        if sound_abbreviation not in count_by_sound:
            count_by_sound[sound_abbreviation] = 0
            filtered_count_by_sound[sound_abbreviation] = 0
        count_by_sound[sound_abbreviation] += 1
        filtered_count_by_sound[sound_abbreviation] += 1 if is_matching_part else 0
        if family_abbreviation not in count_by_family:
            count_by_family[family_abbreviation] = 0
            filtered_count_by_family[family_abbreviation] = 0
        count_by_family[family_abbreviation] += 1
        filtered_count_by_family[family_abbreviation] += 1 if is_matching_part else 0
        if sound not in sounds:
            sounds.append(sound)
        if sound_abbreviation not in sound_abbreviations:
            sound_abbreviations.append(sound_abbreviation)
            if instrumental:
                instrument_abbreviations.append(sound_abbreviation)
        if not instrumental:
            voice_abbreviations.append(sound_abbreviation)
        if family not in families:
            families.append(family)
            family_abbreviations.append(family_abbreviation)
            if instrumental:
                instrumental_family_abbreviations.append(family_abbreviation)

    abbreviated_parts_scoring_order = [
        instr + num
        for instr in cfg.scoring_order
        for num in [""] + ROMAN_NUMERALS_FROM_1_TO_20
    ]
    features = {
        SCORING: ",".join(
            sort_list(abbreviated_parts, abbreviated_parts_scoring_order)
        ),
        SOUND_SCORING: ",".join(sort_list(sound_abbreviations, cfg.scoring_order)),
        INSTRUMENTATION: ",".join(
            sort_list(instrument_abbreviations, cfg.scoring_order)
        ),
        VOICES: ",".join(sort_list(voice_abbreviations, cfg.scoring_order)),
        FAMILY_SCORING: ",".join(
            sort_list(family_abbreviations, cfg.scoring_family_order)
        ),
        FAMILY_INSTRUMENTATION: ",".join(
            sort_list(instrumental_family_abbreviations, cfg.scoring_family_order)
        ),
    }
    for sound in sound_abbreviations:
        features[get_sound_feature(sound, NUMBER_OF_PARTS)] = count_by_sound[sound]
        features[
            get_sound_feature(sound, NUMBER_OF_FILTERED_PARTS)
        ] = filtered_count_by_sound[sound]
        score_data[
            get_sound_feature(sound, NUMBER_OF_FILTERED_PARTS)
        ] = filtered_count_by_sound[sound]

    for family in family_abbreviations:
        features[get_family_feature(family, NUMBER_OF_PARTS)] = count_by_family[family]
        features[
            get_family_feature(family, NUMBER_OF_FILTERED_PARTS)
        ] = filtered_count_by_family[family]
        score_data[
            get_family_feature(family, NUMBER_OF_FILTERED_PARTS)
        ] = filtered_count_by_family[family]

    features[get_score_feature(NUMBER_OF_FILTERED_PARTS)] = len(abbreviated_parts)
    features[get_score_feature(NUMBER_OF_PARTS)] = len(score.parts)
    score_data[get_score_feature(NUMBER_OF_FILTERED_PARTS)] = len(abbreviated_parts)
    score_data[get_score_feature(NUMBER_OF_PARTS)] = len(score.parts)

    return score_features.update(features)



import os
from pathlib import PurePath
from typing import Union

import pandas as pd
from pandas import DataFrame

from musif.common.sort import sort_columns
from musif.config import PostProcessConfiguration
from musif.extract.basic_modules.file_name_generic.constants import ARTIST, TITLE
from musif.extract.basic_modules.scoring.constants import INSTRUMENTATION
from musif.extract.constants import ID, WINDOW_ID
from musif.extract.features.core.constants import FILE_NAME
from musif.extract.features.harmony.constants import (
    HARMONY_AVAILABLE,
    KEY_MODULATORY,
    KEY_PREFIX,
)
from musif.extract.features.prefix import get_part_prefix, get_sound_prefix
from musif.logs import perr, pinfo
from musif.process.constants import PRESENCE
from musif.process.utils import (
    _delete_columns,
    join_keys,
    join_keys_modulatory,
    join_part_degrees,
)


class DataProcessor:
    """Processor class that treats columns and information of a DataFrame

    This operator processes information from a DataFrame or a .csv file.
    It deletes unseful columns for analysis and saves important ones.
    Also saves data in several files in .csv format.
    The main method .process() returns a DataFrame and saves data.
    Requires to have a labels file in ./internal_data directory containing
    each label assigned to each score.
    ...

    Attributes
    ----------
    data : DataFrame
        DataFrame extracted with FeaturesExtractor containing all info.
    info: str
        Path to .csv file or Dataframe containing the information from FeaturesExtractor
    """

    def __init__(self, info: Union[str, DataFrame], *args, **kwargs):
        """
        Parameters
        ----------
        *args:  str
            Could be a path to a .yml file, a PostProcessConfiguration object or a
            dictionary. Length zero or one.
        *kwargs : str
            Key words arguments to construct
        kwargs[info]: Union[str, DataFrame]
            Either a path to a .csv file containing the information either a DataFrame
            object fromm FeaturesExtractor
        """
        self._post_config = PostProcessConfiguration(*args, **kwargs)
        self.info = info
        self.data = self._process_info(self.info)

    def _process_info(self, info: Union[str, DataFrame]) -> DataFrame:
        """
        Extracts the info from a directory to a csv file or from a Dataframe object.

        Parameters
        ------
        info: str
            Info in the from of str (path to csv file) or DataFrame

        Raises
        ------
        FileNotFoundError
            If path to the .csv file is not found.

        Returns
        ------
            Dataframe contaning the information to be processed.
        """

        try:
            if isinstance(info, str) or isinstance(info, PurePath):
                pinfo("\nReading csv file...")
                if not os.path.exists(info):
                    raise FileNotFoundError("A .csv file could not be found")
                if isinstance(info, PurePath):
                    self.destination_route = str(info.with_suffix(""))
                else:
                    self.destination_route = info.replace(".csv", "")
                df = pd.read_csv(
                    info, low_memory=False, sep=",", encoding_errors="replace"
                )
                if df.empty:
                    raise FileNotFoundError("The .csv file could not be found.")
                return df

            elif isinstance(info, DataFrame):
                # pinfo("\nProcessing DataFrame...")
                return info
            else:
                perr(
                    "Wrong info type! You must introduce either a DataFrame either the name of a .csv file."
                )
                return pd.DataFrame()

        except OSError as e:
            perr(
                f"Data could not be loaded. Either wrong path or an empty file was found. {e}"
            )
            return e

    def process(self) -> DataFrame:
        """
        Main method of the class. Removes NaN values, deletes unuseful columns
        and merges those that are needed according to config.yml file.

        Returns
        ------
        Dataframe object
        """

        pinfo("\nPost-processing data...")
        self.data.dropna(axis=1, how="all", inplace=True)
        if self._post_config.delete_files_without_harmony:
            self.delete_files_without_harmony()

        if self._post_config.separate_intrumentation_column:
            pinfo('\nSeparating "Instrumentation" column...')
            self.separate_instrumentation_column()

        self.delete_undesired()

        if self._post_config.grouped_analysis:
            self.group_columns()

        self._final_data_processing()
        return self

    def delete_files_without_harmony(self):
        """
        Deletes files (actually rows in the DataFrame) that didn't have a proper
        harmonic analysis and, there fore, got a value of 0 in 'Harmony_Available'
        column
        """
        if HARMONY_AVAILABLE in self.data:
            number_files = len(self.data[self.data[HARMONY_AVAILABLE] == 0])
            pinfo(
                f"{number_files} files were found without mscx analysis or errors in harmonic analysis. They'll be deleted."
            )
            pinfo(
                f"{self.data[self.data[HARMONY_AVAILABLE] == 0][FILE_NAME].to_string()}"
            )

    def group_columns(self) -> None:
        """
        Groups Key_*_PercentageMeasures, Key_Modulatory and Degrees columns. Into bigger
        groups for agregated analysis, keeping the previous ones. Also deletes
        unnecesary columns for analysis.
        """
        try:
            self._group_keys_modulatory()
            self._group_keys()
            self._join_degrees()
            self._join_degrees_relative()
        except KeyError:
            perr("Some columns to group could not be found.")

    def separate_instrumentation_column(self) -> None:
        """
        Separates Instrumentation column into as many columns as instruments present in
        Instrumentation, assigning a value of 1 for every instrument that is present and
        0 if it is not for every row (aria).
        """
        for i, row in enumerate(self.data[INSTRUMENTATION]):
            for element in row.split(","):
                self.data.at[i, PRESENCE + "_" + element] = 1

        self.data[[i for i in self.data if PRESENCE + "_" in i]] = (
            self.data[[i for i in self.data if PRESENCE + "_" in i]]
            .fillna(0)
            .astype(int)
        )

    def delete_undesired(self, **kwargs) -> None:
        """Deletes not necessary columns and rows for statistical analysis.

        If keyword arguments are passed in, they overwrite those found
        into configurationg file

        Parameters
        ----------
        **kwargs : str, optional
            Any value from config.yml can be overwritten by passing arguments
            to the method

        Raises
        ------
        KeyError
            If any of the columns required to delete is not found
            in the original DataFrame.
        """
        config_data = self._post_config.__dict__
        config_data.update(kwargs)  # Override values

        # deleting columns that are completely nans
        idx = self.data.isna().all()
        to_delete = self.data.columns[idx].to_list()
        self.data.drop(columns=to_delete, inplace=True)

        # deleting rows
        th = config_data["max_nan_rows"] or 1.0
        idx = self.data.isna().sum(axis=1) / self.data.shape[1] > th
        to_delete = self.data.index[idx]
        self.data.drop(index=to_delete, inplace=True)

        # delete cols
        _delete_columns(self.data, config_data)

    def replace_nans(self) -> None:
        # pinfo("Replacing NaN values in selected columns")
        cols = []
        for col in self.data.columns:
            if any(
                substring.lower() in col.lower()
                for substring in tuple(self._post_config.replace_nans)
            ):
                cols.append(col)
        cols = self.data[cols].select_dtypes(include="number").columns
        self.data[cols] = self.data[cols].fillna(0)

    def save(
        self, dest_path: Union[str, PurePath], ext=".csv", ft="csv", **kwargs
    ) -> None:
        """Saves current information into a file given the name of dest_path

        To load one of those file, remember to set the index to
        `musif.extract.constant.ID`, and, if windows are used, to
        `musif.extract.constant.WINDOW_ID`:

        ```python
        df = pd.read_csv('window_alldata.csv').set_index(['Id', 'WindowId'])
        ```

        Parameters
        ----------
        dest_path : str or Path
            Path to directory where the file will be stored; a suffix like
            `_metadata.csv` will be added.
        ext : str
            Extension used to save files. Use `.gz`, `.xz`, `.zip` etc. to compress the
            files. Default: `.csv`
        ft : str
            Type of file for saving. The filetype must be supported by `pandas`, e.g.
            `to_csv`, `to_feather`, `to_parquet`, etc. Default: `csv`
        """

        pinfo(f"Writing data to {dest_path}_*{ext}")
        ft = "to_" + ft
        dest_path = str(dest_path)
        if ft == "csv":
            kwargs["index"] = False
        getattr(self.data, ft)(dest_path + "_alldata" + ext, **kwargs)

    def _group_keys_modulatory(self) -> None:
        self.data.update(
            self.data[
                [i for i in self.data.columns if KEY_PREFIX + KEY_MODULATORY in i]
            ].fillna(0)
        )
        join_keys_modulatory(self.data)

    def _group_keys(self) -> None:
        self.data.update(
            self.data[[i for i in self.data.columns if KEY_PREFIX in i]].fillna(0)
        )
        join_keys(self.data)

    def _join_degrees(self) -> None:
        total_degrees = [
            i for i in self.data.columns if "_Degree" in i and "relative" not in i
        ]

        for part in self._post_config.instruments_to_keep:
            join_part_degrees(total_degrees, get_part_prefix(part), self.data)
        join_part_degrees(total_degrees, get_sound_prefix("voice"), self.data)

    def _join_degrees_relative(self) -> None:
        total_degrees = [
            i for i in self.data.columns if "_Degree" in i and "relative" in i
        ]

        for part in self._post_config.instruments_to_keep:
            join_part_degrees(
                total_degrees, get_part_prefix(part), self.data, sufix="_relative"
            )
        join_part_degrees(
            total_degrees, get_sound_prefix("voice"), self.data, sufix="_relative"
        )

    def _final_data_processing(self) -> None:
        self.data.sort_values([ID, WINDOW_ID], inplace=True)
        self.replace_nans()
        self.data = self.data.reindex(sorted(self.data.columns), axis=1)
        if TITLE and ARTIST in self.data.columns:
            priority_columns = [FILE_NAME, TITLE, ARTIST]
        else:
            priority_columns = []
        self.data = sort_columns(self.data, [ID, WINDOW_ID] + priority_columns)
        self.data.drop("index", axis=1, inplace=True, errors="ignore")



import io
import os
import shutil
import sys
import subprocess
import platform
import zipfile
import requests
from pathlib import Path

from musif.logs import linfo
from tqdm import tqdm

JSYMBOLIC_URL = "https://master.dl.sourceforge.net/project/jmir/jSymbolic/jSymbolic%202.2/jSymbolic_2_2_user.zip"


class JavaNotFoundError(RuntimeError):
    pass


def get_java_path():
    java = __get_java_path_env()
    if java is not None:
        return java
    java = __get_java_path_subprocess()
    if java is not None:
        return java
    java = __get_java_path_env()
    if java is not None:
        return java
    raise JavaNotFoundError(
        "Could not find Java Runtime Environment (JRE)! Please, set JAVA_HOME environment variable or make it available in the PATH."
    )


def __get_java_path_env():
    if sys.platform == "win32":
        java = os.path.join(os.environ["JAVA_HOME"], "bin", "java.exe")
    else:
        java = os.path.join(os.environ["JAVA_HOME"], "bin", "java")
    if os.path.exists(java):
        return java
    else:
        return None


def __get_java_path_shutil():
    path = shutil.which("java")
    return path


def __get_java_path_subprocess():
    try:
        return subprocess.check_output(["which", "java"]).decode("utf-8").strip()
    except subprocess.CalledProcessError:
        return None


def _jsymbolic_path():
    if platform.system() == "Windows":
        cache_dir = Path(os.getenv("APPDATA"), "jSymbolic")
    elif platform.system() == "Linux":
        cache_dir = Path.home() / ".cache" / "jSymbolic"
    elif platform.system() == "Darwin":
        cache_dir = Path.home() / "Library" / "Caches" / "jSymbolic"
    else:
        raise Exception("Unsupported operating system: " + platform.system())
    if not cache_dir.exists():
        cache_dir.mkdir(parents=True)
    filename = cache_dir / "jSymbolic.jar"
    return filename


def download(url):
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    data = b''
    with tqdm(
        desc=url,
        total=total,
        unit='iB',
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for chunk in resp.iter_content(chunk_size=1024):
            data += chunk
            bar.update(len(chunk))
    return data


def download_jsymbolic():
    # Download jSymbolic, if not already downloaded
    # and save it to the OS cache directory
    filename = _jsymbolic_path()
    libdir = filename.parent / 'lib'

    if not filename.exists() or not libdir.exists():
        # Download the zip file
        data = download(JSYMBOLIC_URL)

        # Extract the jar file from the zip file
        with zipfile.ZipFile(io.BytesIO(data)) as zf:

            # Extract the lib directory recursively
            libdir.mkdir()
            for f in zf.filelist:
                if f.filename.startswith('jSymbolic_2_2_user/lib'):
                    with zf.open(f.filename) as fin:
                        with open(libdir.as_posix() + f.filename[len('jSymbolic_2_2_user/lib'):], 'wb') as fout:
                            fout.write(fin.read())

            # Extract the jar file
            with zf.open('jSymbolic_2_2_user/jSymbolic2.jar') as f:
                with open(filename.as_posix(), 'wb') as fout:
                    fout.write(f.read())
        linfo("jSymbolic downloaded into {filname}")



"""Sets the tyoe of encoding to be used"""
ENCODING = "utf-8"

"""Sets the tyoe of encoding to be used"""
CSV_DELIMITER = ","

"""Prefix for voices"""
VOICE_FAMILY = "voice"

GENERAL_FAMILY = "general"



from typing import List

import numpy as np
from pandas import DataFrame

from musif.common._constants import VOICE_FAMILY
from musif.config import ExtractConfiguration
from musif.extract.basic_modules.scoring.constants import FAMILY
from musif.extract.common import _filter_parts_data, _part_matches_filter
from musif.extract.constants import (
    DATA_FAMILY,
    DATA_FAMILY_ABBREVIATION,
    DATA_PART_ABBREVIATION,
    DATA_SOUND_ABBREVIATION,
)
from musif.extract.features.core.constants import NUM_NOTES
from musif.extract.features.core.handler import DATA_NOTES
from musif.extract.features.prefix import get_part_feature, get_part_prefix

from .constants import *


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    parts_data = _filter_parts_data(parts_data, cfg.parts_filter)

    if len(parts_data) == 0:
        return

    features = {}

    for part_data, part_features in zip(parts_data, parts_features):

        part = part_data[DATA_PART_ABBREVIATION]

        features[get_part_feature(part, NUM_NOTES)] = part_features[NUM_NOTES]

    score_features.update(features)

    notes = {}

    for j, part in enumerate(parts_data):

        if part[DATA_PART_ABBREVIATION].startswith("vn"):

            # capitalization to preserve I and II in Violins

            notes[
                part[DATA_PART_ABBREVIATION][0].upper()
                + part[DATA_PART_ABBREVIATION][1:]
            ] = len(part[DATA_NOTES])

        elif part[DATA_FAMILY] == VOICE_FAMILY:

            notes[part[DATA_FAMILY_ABBREVIATION].capitalize()] = int(
                score_features[
                    FAMILY + part[DATA_FAMILY_ABBREVIATION].capitalize() + "_NotesMean"
                ]
            )

        else:

            abbreviation = (
                part[DATA_SOUND_ABBREVIATION][0].upper()
                + part[DATA_SOUND_ABBREVIATION][1:]
            )

            notes[part[DATA_SOUND_ABBREVIATION].capitalize()] = int(
                score_features["Sound" + abbreviation + "_NotesMean"]
            )

    notes = list(notes.items())
    for i in range(len(notes)):

        key_1, value_1 = notes[i]

        for j in range(i + 1, len(notes)):
            key_2, value_2 = notes[j]
            if value_2 == 0:
                if value_1 > 0:
                    texture = np.inf
                else:
                    texture = np.nan
            else:
                texture = value_1 / value_2

            part1_prefix = get_part_prefix(key_1).replace("_", "")

            part2_prefix = get_part_prefix(key_2).replace("_", "")

            score_features[f"{part1_prefix}|{part2_prefix}_{TEXTURE}"] = texture


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):

    if not _part_matches_filter(part_data[DATA_PART_ABBREVIATION], cfg.parts_filter):

        return {}

    notes = part_data[DATA_NOTES]

    part_features.update({DATA_NOTES: len(notes)})



PART_NAME = "PartName"
PART_NUMBER = "PartNumber"
PART_ABBREVIATION = "PartAbbreviation"
SOUND = "Sound"
SOUND_ABBREVIATION = "SoundAbbreviation"
ABBREVIATION = "abbreviation"
FAMILY = "Family"
FAMILY_ABBREVIATION = "FamilyAbbreviation"
INSTRUMENTAL = "Instrumental"
NUMBER_OF_PARTS = "NumberOfParts"
NUMBER_OF_FILTERED_PARTS = "NumberOfFilteredParts"
NUMBER_OF_PARTS_MEAN = "NumberOfPartsMean"
NUMBER_OF_PARTS_STD = "NumberOfPartsStd"
SCORING = "Scoring"
SOUND_SCORING = "SoundScoring"
INSTRUMENTATION = "Instrumentation"
VOICES = "Voices"
FAMILY_SCORING = "FamilyScoring"
FAMILY_INSTRUMENTATION = "FamilyInstrumentation"
ROLE_TYPE = "RoleType"



from typing import List

from musif.config import ExtractConfiguration
from musif.extract.features.core.handler import DATA_KEY, DATA_MODE, DATA_KEY_NAME
from musif.musicxml.key import (
    _get_key_signature,
    get_key_signature_type,
    get_name_from_key,
)
from .constants import *


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    key = score_data[DATA_KEY]
    key_name = score_data[DATA_KEY_NAME]
    mode = score_data[DATA_MODE]
    # key_signature = get_key_signature(key)
    key_signature = score_data[KEY_SIGNATURE]
    key_signature_type = get_key_signature_type(key_signature)

    score_features.update(
        {
            KEY: key_name,
            # KEY_SIGNATURE: key_signature,
            KEY_SIGNATURE_TYPE: key_signature_type,
            MODE: mode,
        }
    )


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    pass



from os import path
from typing import List
import pandas as pd
from music21 import *
from music21.stream import Measure
from musif.extract.constants import DATA_FILTERED_PARTS
from musif.extract.features.tempo.constants import TIME_SIGNATURE

from musif.config import ExtractConfiguration
from musif.extract.basic_modules.scoring.constants import (
    FAMILY_ABBREVIATION,
    NUMBER_OF_FILTERED_PARTS,
    SOUND_ABBREVIATION,
)
from musif.extract.constants import (
    DATA_FAMILY_ABBREVIATION,
    DATA_FILE,
    DATA_MUSESCORE_SCORE,
    DATA_PART,
    DATA_PART_ABBREVIATION,
    DATA_SCORE,
    DATA_SOUND_ABBREVIATION,
    GLOBAL_TIME_SIGNATURE,
)
from musif.extract.features.prefix import (
    get_family_feature,
    get_part_feature,
    get_score_feature,
    get_sound_feature,
)
from musif.musicxml.common import (
    _get_intervals,
    _get_lyrics_in_notes,
    get_notes_and_measures,
)
from musif.musicxml.key import get_key_and_mode, _get_key_signature, get_name_from_key

from .constants import *


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    part = part_data[DATA_PART]
    (
        notes,
        measures,
        sounding_measures,
        notes_and_rests,
    ) = get_notes_and_measures(part)
    lyrics = _get_lyrics_in_notes(notes)
    intervals = _get_intervals(notes)
    part_data.update(
        {
            DATA_NOTES: notes,
            DATA_LYRICS: lyrics,
            DATA_SOUNDING_MEASURES: sounding_measures,
            DATA_MEASURES: measures,
            DATA_INTERVALS: intervals,
            DATA_NOTES_AND_RESTS: notes_and_rests,
        }
    )
    part_features.update(
        {
            NUM_NOTES: len(notes),
            NUM_MEASURES: len(measures),
            NUM_SOUNDING_MEASURES: len(sounding_measures),
        }
    )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    score = score_data[DATA_SCORE]
    score_key, key_name, mode = get_key_and_mode(score)
    if (
        cfg.is_requested_musescore_file()
        and (score_data[DATA_MUSESCORE_SCORE] is not None)
        and (not score_data[DATA_MUSESCORE_SCORE].empty)
    ):
        tonality_ms3 = score_data[DATA_MUSESCORE_SCORE].globalkey[0]
        if key_name != tonality_ms3:
            score_key = key.Key(tonality_ms3)
            mode, key_name = get_name_from_key(score_key)

    score_features[FILE_NAME] = path.abspath(score_data[DATA_FILE])
    num_measures = len(score.parts[0].getElementsByClass(Measure))
    key_signature = _get_key_signature(score_key)

    part = score.parts[0]
    time_signature = _get_time_signature(score_data)

    score_data.update(
        {
            DATA_KEY: score_key,
            KEY_SIGNATURE: key_signature,
            TIME_SIGNATURE: time_signature,
            KEY_SIGNATURE: key_signature,
            DATA_KEY_NAME: key_name,
            DATA_MODE: mode,
            DATA_MEASURES: num_measures,
        }
    )

    features = {}
    for i, part in enumerate(parts_features):
        part[SOUND_ABBREVIATION] = parts_data[i][DATA_SOUND_ABBREVIATION]
        part[FAMILY_ABBREVIATION] = parts_data[i][DATA_FAMILY_ABBREVIATION]

    df_parts = pd.DataFrame(parts_features)
    df_sound = df_parts.groupby(SOUND_ABBREVIATION).aggregate(
        {NUM_NOTES: "sum", NUM_SOUNDING_MEASURES: "sum"}
    )
    df_family = df_parts.groupby(FAMILY_ABBREVIATION).aggregate(
        {NUM_NOTES: "sum", NUM_SOUNDING_MEASURES: "sum"}
    )
    df_score = df_parts.aggregate({NUM_NOTES: "sum", NUM_SOUNDING_MEASURES: "sum"})

    for part_data, part_features in zip(parts_data, parts_features):
        part = part_data[DATA_PART_ABBREVIATION]
        features[get_part_feature(part, NUM_NOTES)] = part_features[NUM_NOTES]
        features[get_part_feature(part, NUM_SOUNDING_MEASURES)] = part_features[
            NUM_SOUNDING_MEASURES
        ]
        part_data[get_part_feature(part, NUM_NOTES)] = part_features[NUM_NOTES]
        part_data[get_part_feature(part, NUM_SOUNDING_MEASURES)] = part_features[
            NUM_SOUNDING_MEASURES
        ]

    for sound in df_sound.index:
        notes = df_sound.loc[sound, NUM_NOTES].tolist()
        sounding_measures = df_sound.loc[sound, NUM_SOUNDING_MEASURES].tolist()
        sound_parts = score_data[get_sound_feature(sound, NUMBER_OF_FILTERED_PARTS)]
        notes_mean = notes / sound_parts if sound_parts > 0 else 0
        sounding_measures_mean = (
            sounding_measures / sound_parts if sound_parts > 0 else 0
        )
        features[get_sound_feature(sound, NUM_NOTES)] = notes
        features[get_sound_feature(sound, NOTES_MEAN)] = notes_mean
        features[get_sound_feature(sound, NUM_SOUNDING_MEASURES)] = sounding_measures
        features[
            get_sound_feature(sound, SOUNDING_MEASURES_MEAN)
        ] = sounding_measures_mean

    for family in df_family.index:
        notes = df_family.loc[family, NUM_NOTES].tolist()
        sounding_measures = df_family.loc[family, NUM_SOUNDING_MEASURES].tolist()
        family_parts = score_data[get_family_feature(family, NUMBER_OF_FILTERED_PARTS)]
        notes_mean = notes / family_parts if family_parts > 0 else 0
        sounding_measures_mean = (
            sounding_measures / family_parts if family_parts > 0 else 0
        )
        features[get_family_feature(family, NUM_NOTES)] = notes
        features[get_family_feature(family, NOTES_MEAN)] = notes_mean
        features[get_family_feature(family, NUM_SOUNDING_MEASURES)] = df_family.loc[
            family, NUM_SOUNDING_MEASURES
        ].tolist()
        features[
            get_family_feature(family, SOUNDING_MEASURES_MEAN)
        ] = sounding_measures_mean

    notes = df_score[NUM_NOTES].tolist()
    notes_mean = notes / len(parts_data)
    sounding_measures = df_score[NUM_SOUNDING_MEASURES].tolist()
    sounding_measures_mean = sounding_measures / len(parts_data)
    features[get_score_feature(NUM_NOTES)] = notes
    features[get_score_feature(NOTES_MEAN)] = notes_mean
    features[get_score_feature(NUM_SOUNDING_MEASURES)] = sounding_measures
    features[get_score_feature(SOUNDING_MEASURES_MEAN)] = sounding_measures_mean
    features[NUM_MEASURES] = num_measures
    score_features.update(features)


def _get_time_signature(score_data: dict) -> str:
    """
    This function takes in a dictionary of score data and returns the global time
    signature of the score as a string.

    Parameters:
    score_data (dict): A dictionary containing score data.

    Returns:
    str: A string representing the time signature of the score. It is `'NA'`
        if no time signature is found.
    """
    if hasattr(score_data.get(GLOBAL_TIME_SIGNATURE), 'ratioString'):
        time_signature = score_data[GLOBAL_TIME_SIGNATURE].ratioString
    else:
        first_measure = score_data[DATA_FILTERED_PARTS][0].getElementsByClass(Measure)[
            0
        ]
        if hasattr(first_measure.timeSignature, 'ratioString'):
            time_signature = first_measure.timeSignature.ratioString
        else:
            time_signature = "NA"
    return time_signature



from statistics import mean
from typing import List, Optional, Tuple

from music21.chord import Chord
from music21.harmony import ChordSymbol

from music21.note import Note

from musif.config import ExtractConfiguration
from musif.extract.common import _filter_parts_data
from musif.extract.constants import DATA_PART_ABBREVIATION
from musif.extract.features.core.constants import DATA_NOTES
from musif.cache import isinstance

from ..prefix import get_part_feature
from .constants import *


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    notes = part_data[DATA_NOTES]
    if notes is None or len(notes) == 0:
        return
    lowest_note, highest_note = _get_notes_ambitus(notes)
    lowest_note_text = lowest_note.nameWithOctave.replace("-", "b")
    highest_note_text = highest_note.nameWithOctave.replace("-", "b")
    lowest_note_index = int(lowest_note.pitch.midi)
    highest_note_index = int(highest_note.pitch.midi)
    total_ambitus = highest_note_index - lowest_note_index
    ambitus_features = {
        LOWEST_NOTE: lowest_note_text,
        HIGHEST_NOTE: highest_note_text,
        LOWEST_NOTE_INDEX: lowest_note_index,
        HIGHEST_NOTE_INDEX: highest_note_index,
        AMBITUS: total_ambitus,
    }
    part_features.update(ambitus_features)


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):

    parts_data = _filter_parts_data(parts_data, cfg.parts_filter)
    if len(parts_data) == 0:
        return

    for part_data, part_features in zip(parts_data, parts_features):
        part = part_data[DATA_PART_ABBREVIATION]
        for feature_name in SCORE_FEATURES:
            part_feature = get_part_feature(part, feature_name)
            if feature_name in score_features:
                score_features[part_feature] = mean(
                    [score_features[part_feature], parts_features[feature_name]]
                )
            else:
                score_features[part_feature] = part_features.get(feature_name)


def _get_notes_ambitus(notes: List[Note]) -> Tuple[Note, Note]:
    first_note = notes[0][0] if isinstance(notes[0], (Chord, ChordSymbol)) else notes[0]
    lowest_note = first_note
    highest_note = first_note
    for note in notes[1:]:
        current_note = note[0] if isinstance(note, (Chord, ChordSymbol)) else note
        if current_note.pitch.midi < lowest_note.pitch.midi:
            lowest_note = current_note
        if current_note.pitch.midi > highest_note.pitch.midi:
            highest_note = current_note
    return lowest_note, highest_note



from typing import List, Optional, Tuple

from music21.stream.base import Part
from roman import fromRoman, toRoman

from musif.config import ExtractConfiguration

ROMAN_NUMERALS_FROM_1_TO_20 = [toRoman(i).upper() for i in range(1, 21)]



def to_abbreviation(part: Part, parts: List[Part], cfg: ExtractConfiguration) -> str:
    """
        Returns abbreviation name for a specific part based on the sound name
        
        Parameters
        ----------
        part:  str
            Part to get abbreviaton from
        parts: List[part]
            List of parts in the score
        cfg: ExtractConfiguration
            ExtractConfiguration object            
    """
    
    sound = extract_sound(part, cfg)
    return list(_extract_abbreviated_part(sound, part, parts, cfg))[0]


def extract_sound(part: Part, config: ExtractConfiguration) -> str:
    """
    Returns sound name for a specific part based on the sound name
    Part: VnI -> Sound name: Vn
    Parameters
    ----------
    part:  str
        Part to get abbreviaton from
    cfg: ExtractConfiguration
        ExtractConfiguration object            
    """
    
    instrument = part.getInstrument(returnDefault=False)
    sound_name = None
    if instrument is None or instrument.instrumentSound is None:
        sound_name = part.partName.strip().split(' ')[0]
        if sound_name not in config.sound_to_abbreviation:
            sound_name = _replace_naming_exceptions(sound_name, part)
            sound_name = sound_name if sound_name[-1] != 's' else sound_name[:-1]
    else:
        for instrument in instrument.instrumentSound.split(".")[::-1]:
            if (
                "flat" not in instrument
                and "sharp" not in instrument
                and len(instrument) > 2
            ):
                sound_name = instrument
                break
    if sound_name:
        sound_name = sound_name.replace('-', ' ').lower()
        sound_name = _replace_naming_exceptions(sound_name, part)
        return sound_name
    else:
        return ''


def _extract_abbreviated_part(sound: str, part: Part, parts: List[Part], config: ExtractConfiguration) -> Tuple[str, str, int]:
    if sound not in config.sound_to_abbreviation:
        abbreviation = part.partAbbreviation  # may contain I, II or whatever
        abbreviation_parts = abbreviation.split(" ")
        abbreviation = abbreviation_parts[0].lower().replace(".", "") + (
            abbreviation_parts[1] if len(abbreviation_parts) > 1 else ""
        )
    else:
        abbreviation = config.sound_to_abbreviation[sound]
    part_roman_number = _get_part_roman_number(
        part) or _get_part_roman_number_by_position(part, parts)
    other_number = _get_part_normal_number(part)
    part_number = fromRoman(part_roman_number) if part_roman_number else (other_number if other_number else 0)
    sound_abbreviation = abbreviation.split('(')[0]
    return abbreviation + (part_roman_number or ""), sound_abbreviation, part_number

def _get_part_normal_number(part: Part) -> Optional[str]:
    # this is didone-specific
    if '(' and ')' in part.partAbbreviation:
        return part.partAbbreviation.split('(')[1].split(')')[0]
    return None

def _get_part_roman_number(part: Part) -> Optional[str]:
    for number in ROMAN_NUMERALS_FROM_1_TO_20:
        if part.partAbbreviation.endswith(f". {number}") or part.partName.endswith(
            f" {number}"
        ):
            return number
    return None


def _get_part_roman_number_by_position(part: Part, parts: List[Part]) -> Optional[str]:
    same_sound_parts = [same_sound_part
                        for same_sound_part in parts
                        if part.getInstrument(returnDefault=False) == same_sound_part.getInstrument(returnDefault=False)]
    if len(same_sound_parts) > 1:
        for i, same_sound_part in enumerate(same_sound_parts, 1):
            if part.partName == same_sound_part.partName:
                return toRoman(i)
    return None


def _replace_naming_exceptions(sound: str, part: Part) -> str:
    if 'da caccia' in sound:
        sound = sound.replace('da caccia', '')
        if 'tromba' in sound:
            sound = 'horn'
    if 'bass' == sound:  # determines if voice or string instrument
        if len(part.lyrics()) == 0:
            sound = "basso continuo"
    if "french" in sound and "horn" in sound:
        sound = "horn"
    if "cello" in sound and "bass" in part.partName.lower():
        sound = "basso continuo"
    return sound



LOWEST_NOTE = "LowestNote"
HIGHEST_NOTE = "HighestNote"
LOWEST_NOTE_INDEX = "LowestNoteIndex"
HIGHEST_NOTE_INDEX = "HighestNoteIndex"
AMBITUS = "Ambitus"
SCORE_FEATURES = [
    LOWEST_NOTE_INDEX,
    LOWEST_NOTE,
    HIGHEST_NOTE_INDEX,
    HIGHEST_NOTE,
    AMBITUS,
]



from enum import Enum
from typing import List

from music21.expressions import TextExpression
from music21.stream import Measure

from musif.cache import isinstance
from musif.config import ExtractConfiguration
from musif.extract.constants import (
    DATA_NUMERIC_TEMPO,
    DATA_PART,
    DATA_SCORE,
    GLOBAL_TIME_SIGNATURE,
)
from musif.musicxml.tempo import (
    get_number_of_beats,
    get_tempo_grouped_1,
    get_tempo_grouped_2,
    get_time_signature_type,
)

from . import constants as C


class TempoGroup2(Enum):
    NA = "NA"
    SLOW = "Slow"
    MODERATE = "Moderate"
    FAST = "Fast"


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    part = part_data[DATA_PART]

    (
        time_signature,
        measures,
        time_signatures,
        time_signature_grouped,
        number_of_beats,
    ) = extract_time_signatures(list(part.getElementsByClass(Measure)), score_data)
    part_data.update(
        {
            C.TIME_SIGNATURES: time_signatures,
            C.TS_MEASURES: measures,
        }
    )
    part_features.update(
        {
            C.TIME_SIGNATURE: time_signature,
            C.TIME_SIGNATURE_GROUPED: time_signature_grouped,
            C.NUMBER_OF_BEATS: number_of_beats,
        }
    )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):

    score = score_data[DATA_SCORE]
    # for the features, we use the first part as reference!
    part = score.parts[0]
    numeric_tempo, tempo_mark = extract_tempo(score_data, part)
    tempo_grouped_1 = get_tempo_grouped_1(tempo_mark)
    tempo_grouped_2 = get_tempo_grouped_2(tempo_grouped_1)

    (
        time_signature,
        measures,
        time_signatures,
        time_signature_grouped,
        number_of_beats,
    ) = extract_time_signatures(list(part.getElementsByClass(Measure)), score_data)

    score_features.update(
        {
            C.TEMPO: tempo_mark,
            C.NUMERIC_TEMPO: numeric_tempo,
            C.TIME_SIGNATURE: time_signature,
            C.TIME_SIGNATURE_GROUPED: time_signature_grouped,
            C.NUMBER_OF_BEATS: number_of_beats,
            C.TEMPO_GROUPED_1: tempo_grouped_1,
            C.TEMPO_GROUPED_2: tempo_grouped_2,
        }
    )


def extract_time_signatures(measures: list, score_data: dict) -> tuple:
    """
    Extracts time signatures from a list of measures and returns relevant information.

    Args:
        measures (list): A list of measures.
        score_data (dict): A dictionary containing score data.

    Returns:
        tuple: A tuple containing the time signature, dictionary of measures and their
        indices, list of time signatures, time signature type, and number of beats.

    The function extracts time signatures from a list of measures and returns
    relevant information. The function takes two arguments, measures and
    score_data. measures is a list of measures and score_data is a dictionary
    containing score data. The function returns a tuple containing the time
    signature, dictionary of measures and their indices, list of time signatures,
    time signature type, and number of beats.

    The function ases `'NA'` as the time signature if no time signature is found
    in the score.
    """

    ts_measures = {}
    time_signatures = []
    for i, element in enumerate(measures):
        if element.measureNumber not in ts_measures:
            if element.timeSignature is not None:
                time_signatures.append(element.timeSignature.ratioString)
            else:
                if len(time_signatures) >= 1:
                    time_signatures.append(time_signatures[-1])
                elif hasattr(score_data.get(GLOBAL_TIME_SIGNATURE), 'ratioString'):
                    time_signatures.append(
                        score_data[GLOBAL_TIME_SIGNATURE].ratioString
                    )
                else:
                    time_signatures.append("NA")

            ts_measures[element.measureNumber] = i
        else:
            time_signatures.append(time_signatures[ts_measures[element.measureNumber]])

    time_signatures_set = set(time_signatures)
    time_signature = list(sorted(time_signatures_set, key=time_signatures.index))[0]

    time_signature_grouped = get_time_signature_type(time_signature)
    number_of_beats = get_number_of_beats(time_signature)

    return (
        time_signature,
        ts_measures,
        time_signatures,
        time_signature_grouped,
        number_of_beats,
    )


def extract_tempo(score_data, part):
    numeric_tempo = score_data[DATA_NUMERIC_TEMPO]
    tempo_mark = TempoGroup2.NA.value
    for measure in part:
        if isinstance(measure, Measure):
            for element in measure:
                if isinstance(element, TextExpression):
                    if get_tempo_grouped_1(element.content) != "NA":
                        tempo_mark = element.content
            break  # only take into account the first bar!
    return numeric_tempo, tempo_mark



def get_gender(character: str) -> str:
    """
    Returns gender of a character depending on it's name.
    """

    if character in ["Didone", "Selene", "Dircea", "Creusa", "Semira", "Mandane"]:
        return "female"
    else:
        return "male"


def get_role(character: str) -> str:
    """
    Returns general role type for specific operatic characters
    """

    if character in ["Demofoonte", "Licomede", "Tito", "Catone", "Fenicio"]:
        return "Senior ruler"
    elif character in [
        "Didone",
        "Dircea",
        "Cleofide",
        "Mandane",
        "Deidamia",
        "Sabina",
        "Vitellia",
        "Marzia",
        "Cleonice",
    ]:
        return "Female lover 1"
    elif character in [
        "Enea",
        "Poro",
        "Arbace",
        "Timante",
        "Achille",
        "Adriano",
        "Sesto",
        "Cesare",
        "Alceste",
        "Demetrio",
    ]:
        return "Male lover 1"
    elif character in [
        "Selene",
        "Creusa",
        "Erissena",
        "Semira",
        "Emirena",
        "Servila",
        "Emilia",
        "Barsene",
    ]:
        return "Female lover 2"
    elif character in [
        "Iarba",
        "Alessandro",
        "Artaserse",
        "Cherinto",
        "Teagene",
        "Farnaspe",
        "Annio",
        "Olinte",
    ]:
        return "Male lover 2"
    elif character in ["Gandarte", "Aquilio", "Fulvio"]:
        return "Male lover 3"
    elif character in ["Araspe", "Megabise", "Adrasto", "Arcade", "Publio", "Mitrano"]:
        return "Confidant"
    elif character in ["Osmida", "Timagene", "Artabano", "Matusio", "Ulisse", "Osroa"]:
        return "Antagonist"



import warnings
from logging import getLogger
from typing import List, Dict

from musif.config import LOGGER_NAME
from pandas import DataFrame


def sort_list(list_to_sort: List[str], reference_list: List[str]) -> List[str]:
    """
    Sorts first list based on the second one. Those elements that are not present
    in the reference list will be placed at the end. Returns the same list re-ordered.

    Parameters
    ----------
    list_to_sort : list
        List that needs to be sorted according to some criteria
    reference_list : str
        List used as reference to sort the first one.
    """
    sort_dictionary = {elem: i for i, elem in enumerate(reference_list)}
    found = [elem for elem in list_to_sort if elem in sort_dictionary]
    orphans = [elem for elem in list_to_sort if elem not in sort_dictionary]
    return sorted(found, key=lambda x: sort_dictionary[x]) + orphans


def sort_dict(dict_to_sort: dict, reference_list: list) -> dict:
    """
    Sorts dictionary keys according to a reference li()

    Parameters
    ----------
    dict_to_sort : dict
        Dictionary that needs to be sorted according to some criteria
    main_list : list
        List used as reference to sort the first one.
    """

    indexes = []
    oprhans = []
    for i in dict_to_sort:
        if i in reference_list:
            indexes.append(reference_list.index(i))
        else:
            oprhans.append({i: dict_to_sort[i]})
            getLogger(LOGGER_NAME).warning(
                "We do not have the appropiate sorting for {}".format(i)
            )
    indexes = sorted(indexes)
    list_sorted = [
        {reference_list[i]: dict_to_sort[reference_list[i]]} for i in indexes
    ]
    list_sorted = list_sorted + oprhans
    dict_sorted = dict((key, d[key]) for d in list_sorted for key in d)
    return dict_sorted


def sort_columns(data: DataFrame, sorting_list: list) -> DataFrame:
    """
    Reorders columns of a Dataframe according to a reference list. Uses sort_list.

    Parameters
    ----------
    data : DataFrame
        DataFrame whose columns needs to be re-ordered according to some criteria
    sorting_list : list
        List used as reference to sort columns.
    """

    cols = sort_list(data.columns.tolist(), [i for i in sorting_list])
    data = data[cols]
    return data


def sort_dataframe(
    data: DataFrame, column: str, sorting_lists: Dict[list, str], key_to_sort: str
) -> DataFrame:
    """
    Sorts Dataframe's rows by a column using a list as a reference.
    Parameters
    ----------
    data: DataFrame
        DataFrame to be re-ordered according to some criteria.
    column: str
        Column of the Dataframe used as key.
    sorting_lists: Dict[list]
        Dictionary containing lists used as reference to sort values.
    key_to_sort: str
        Key from sorting_lists that contains the desired list to be used as reference.
    """
    if key_to_sort == "Alphabetic":
        dataSorted = data.sort_values(by=[column])
    else:
        form_list = sorting_lists[key_to_sort]  # es global
        indexes = []
        for i in data[column]:
            if str(i).lower().strip() not in ["nan", "nd"]:
                value = (
                    i.strip()
                    if key_to_sort not in ["FormSorting", "CharacterSorting"]
                    else i.strip().lower()
                )
                try:
                    # index = form_list.index(value)
                    index = form_list.index(value) if value in form_list else 999
                except ValueError:
                    index = 999
                    warnings.warn(
                        "We do not have the value {} in the sorting list {}".format(
                            value, key_to_sort
                        )
                    )
                indexes.append(index)
            else:
                indexes.append(999)  # at the end of the list

        data.loc[:, "Ranks"] = indexes
        dataSorted = data.sort_values(by=["Ranks"])
        dataSorted.drop("Ranks", 1, inplace=True)
    return dataSorted



DATA_NOTES = "notes"
DATA_NOTES_AND_RESTS = "notes_and_rests"
DATA_LYRICS = "lyrics"
DATA_SOUNDING_MEASURES = "sounding_measures"
DATA_MEASURES = "measures"
DATA_INTERVALS = "intervals"
DATA_KEY = "key"
DATA_KEY_NAME = "key_name"
DATA_MODE = "mode"
FILE_NAME = "FileName"
NUM_NOTES = "Notes"
NOTES_MEAN = "NotesMean"
NUM_SOUNDING_MEASURES = "SoundingMeasures"
SOUNDING_MEASURES_MEAN = "SoundingMeasuresMean"
NUM_MEASURES = "Measures"
KEY_SIGNATURE = "KeySignature"



import logging

"""
Common constants to be used transversaly throught modules
"""

"""Sequences needed to get colored ouput and reset back to normal"""
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"


"""Logging constants to introduce in the logger"""
LEVEL_DEBUG = logging.getLevelName(logging.DEBUG)
LEVEL_INFO = logging.getLevelName(logging.INFO)
LEVEL_WARNING = logging.getLevelName(logging.WARNING)
LEVEL_ERROR = logging.getLevelName(logging.ERROR)
LEVEL_CRITICAL = logging.getLevelName(logging.CRITICAL)

"""Indexes for each color"""
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

"""Dictionary to access different color indexes for logging"""
COLORS = {
    LEVEL_DEBUG: BLUE,
    LEVEL_INFO: WHITE,
    LEVEL_WARNING: YELLOW,
    LEVEL_ERROR: RED,
    LEVEL_CRITICAL: YELLOW,
}



MUSIC21_FILE_EXTENSIONS = [".xml", ".mxl", ".musicxml"]
"""Extensions used by music21. Defaults to `[".xml", ".mxl", ".musicxml"]`"""


SOUND_TO_ABBREVIATION = {
    "accordion": "acc",
    "alto flute": "afl",
    "alto": "alt",
    "voice": "v",
    "contralto": "alt",
    "arpeggione": "arp",
    "bagpipe": "bag",
    "baritone": "bar",
    "bass": "bass",
    "bass baritone": "bbar",
    "continuo": "bs",
    "basso continuo": "bs",
    "continuous bass": "bs",
    "bass clarinet": "bcl",
    "bell": "bell",
    "chimes": "bell",
    "bells": "bell",
    "bass flute": "bfl",
    "bass guitar": "bgtr",
    "electric bass": "bgtr",
    "banjo": "bjo",
    "bassoon": "bn",
    "fagoto": "bn",
    "bass oboe": "bob",
    "baritone oboe": "bob",
    "oboe baritono": "bob",
    "brass instruments": "br",
    "strumento a fiato di ottone": "br",
    "baryton": "bryt",
    "baritono": "bryt",
    "basset clarinet": "bstcl",
    "clarinetto di bassetto": "bstcl",
    "basset horn": "bsthn",
    "corno di bassetto": "bsthn",
    "bugle": "bug",
    "corna de segnale": "bug",
    "contrabass clarinet": "cbcl",
    "clarinetto contrabbasso": "cbcl",
    "contrabassoon": "cbn",
    "contrafagotto": "cbn",
    "children's chorus": "cch",
    "coro di fanciulli": "cch",
    "celesta": "cel",
    "mixed chorus": "ch",
    "coro miste": "ch",
    "cimbalom": "cimb",
    "cittern": "cit",
    "cetra": "cit",
    "cetara": "cit",
    "clarinet": "cl",
    "clarinetto": "cl",
    "clarinetti": "cl",
    "clavichord": "clvd",
    "clavicordo": "clvd",
    "chalumeau": "cm",
    "salmoè": "cm",
    "concertina": "conc",
    "crumhorn": "crh",
    "cromorno": "crh",
    "cornet": "crt",
    "cornetta": "crt",
    "a pistoni": "crt",
    "cornett": "crtt",
    "zink": "crtt",
    "cornetto": "crtt",
    "child's voice": "cv",
    "voce di fanciullo": "cv",
    "double bass": "db",
    "contrabasso": "db",
    "dulcian": "dlcn",
    "dulciana": "dlcn",
    "domra": "dom",
    "dulcimer": "dulc",
    "salterio": "dulc",
    "electric guitar": "egtr",
    "chitarra elettrica": "egtr",
    "english horn": "eh",
    "cor anglais": "eh",
    "corno inglese": "eh",
    "electronic instruments": "elec",
    "strumento elettrica": "elec",
    "electric piano": "epf",
    "pianoforte elettrico": "epf",
    "equal voices": "eq",
    "voci pari": "eq",
    "erhu": "erhu",
    "euphonium": "euph",
    "eufonio": "euph",
    "female chorus": "fch",
    "coro femminile": "fch",
    "flute d'amore": "fda",
    "tenor flute": "fda",
    "flauto tenor": "fda",
    "flugelhorn": "fgh",
    "flicorno soprano": "fgh",
    "fife": "fife",
    "fiffaro": "fife",
    "flute": "fl",
    "flauto": "fl",
    "flageolet": "flag",
    "flagioletto": "flag",
    "glass harmonica": "ghca",
    "bowl organ": "ghca",
    "glassarmonica": "ghca",
    "bicchieri de vetro": "ghca",
    "glockenspiel": "gl",
    "campanelli": "gl",
    "metallofono": "gl",
    "guitar": "gtr",
    "chitarra": "gtr",
    "harmonium": "harm",
    "armonium": "harm",
    "harmonica": "hca",
    "mouth organ": "hca",
    "armonica a bocco": "hca",
    "heckelphone": "heck",
    "heckelfono": "heck",
    "horn": "hn",
    "corno": "hn",
    "harp": "hp",
    "arpa": "hp",
    "harpsichord": "hpd",
    "cembalo": "hpd",
    "clavicembalo": "hpd",
    "keyboard instrument": "kbd",
    "strumento a tastiera": "kbd",
    "lute": "lute",
    "liuto": "lute",
    "lyre": "lyre",
    "lira": "lyre",
    "mandolin": "mand",
    "mandolino": "mand",
    "marimba": "mar",
    "male chorus": "mch",
    "coro maschile": "mch",
    "mezzo-soprano": "mez",
    "mezzosoprano": "mez",
    "musette": "mus",
    "musette de cour": "mus",
    "narrator": "nar",
    "reciter": "nar",
    "narratore": "nar",
    "oboe": "ob",
    "ocarina": "oca",
    "oboe d'amore": "oda",
    "ondes martenot": "om",
    "onde martenot": "om",
    "ophicleide": "oph",
    "orchestra": "orch",
    "organ": "org",
    "organo": "org",
    "oud": "oud",
    "pan flute": "pan",
    "pan-pipes": "pan",
    "flauto di pan": "pan",
    "siringa": "pan",
    "percussion": "perc",
    "strumento a percussione": "perc",
    "piano": "pf",
    "pianoforte": "pf",
    "piano 3 hands": "pf3h",
    "pianoforte 3 mani": "pf3h",
    "piano 4 hands": "pf4h",
    "pianoforte 4 mani": "pf4h",
    "piano 5 hands": "pf5h",
    "pianoforte 5 mani": "pf5h",
    "piano 6 hands": "pf6h",
    "pianoforte 6 mani": "pf6h",
    "piano left hand": "pflh",
    "piano di sinistra": "pflh",
    "pedal piano": "pfped",
    "pianoforte organistico": "pfped",
    "piano right hand": "pfrh",
    "piano di destra": "pfrh",
    "piccolo": "picc",
    "flauto piccolo": "picc",
    "ottavino": "picc",
    "pipa": "pipa",
    "timpani": "timp",
    "piccolo trumpet": "ptpt",
    "trombino": "ptpt",
    "rebec": "reb",
    "rebeca": "reb",
    "ribeca": "reb",
    "recorder": "rec",
    "flauto dolce": "rec",
    "flauto a becco": "rec",
    "sarrusophone": "sar",
    "sarrusofono": "sar",
    "saxophone": "sax",
    "sassofono": "sax",
    "saxofono": "sax",
    "sheng": "sheng",
    "shawm": "shw",
    "ciaramella": "shw",
    "piffaro": "shw",
    "sitar": "sit",
    "sackbut": "skbt",
    "soprano": "sop",
    "serpent": "srp",
    "serpentone": "srp",
    "slide trumpet": "stpt",
    "tromba a tirarsi": "stpt",
    "string instruments": "str",
    "archi": "str",
    "saxhorn": "sxh",
    "flicorni": "sxh",
    "synthesizer": "syn",
    "sintetizzatore": "syn",
    "tuba": "tba",
    "trombone": "tbn",
    "tenor": "ten",
    "tenore": "ten",
    "theremin": "thrm",
    "trumpet": "tpt",
    "tromba": "tpt",
    "clarino": "tpt",
    "unison chorus": "uch",
    "coro unisono": "uch",
    "ukelele": "uke",
    "ukulele": "uke",
    "solo": "v",
    "voce": "v",
    "viola": "va",
    "viola pomposa": "vap",
    "cello": "vc",
    "violoncello": "vc",
    "viola d'amore": "vda",
    "vibraphone": "vib",
    "vibrafono": "vib",
    "vielle": "vie",
    "hurdy-gurdy": "vie",
    "lyra tedesca": "vie",
    "ghironda": "vie",
    "viol": "viol",
    "viola da gamba": "viol",
    "gamba": "viol",
    "violone": "vlne",
    "violin": "vn",
    "violino": "vn",
    "vuvuzela": "vuv",
    "voices": "vv",
    "multiple soloists": "vv",
    "voci": "vv",
    "wagner tuba": "wag",
    "tuba di wagner": "wag",
    "woodwind instruments": "ww",
    "strumento a fiato di legno": "ww",
    "xiao": "xiao",
    "xylophone": "xyl",
    "xilofono": "xyl",
    "silofono": "xyl",
    "zither": "zith",
    "cetra da tavola": "zith",
}
"""Dictionary connecting each instrument name in a musicxml to an abbreviation"""

FAMILY_TO_ABBREVIATION = {
    "strings": "str",
    "woodwinds": "ww",
    "wind": "ww",
    "brass": "br",
    "percussion": "perc",
    "voice": "voice",
    "general": "gen",
}
"""Dictionary containing abbreviation of each family"""

SOUND_TO_FAMILY = {
    "soprano": "voice",
    "cantus": "voice",
    "discant": "voice",
    "discantus": "voice",
    "tiple": "voice",
    "mezzo soprano": "voice",
    "alto": "voice",
    "contralto": "voice",
    "tenor": "voice",
    "baritone": "voice",
    "bass": "voice",
    "voice": "voice",
    "solo vocal part": "voice",
    "dessus": "voice",
    "spoken voice": "voice",
    "highest voice": "voice",
    "quinta vox": "voice",
    "5th voice": "voice",
    "sexta vox": "voice",
    "6th voice": "voice",
    "vagans": "voice",
    "choir": "voice",
    "chorus": "voice",
    "the soprano part in a chorus": "voice",
    "the alto part in a chorus": "voice",
    "the tenor part in a chorus": "voice",
    "the bass part in a chorus": "voice",
    "an unspecified or unknown chorus part": "voice",
    "children's choir": "voice",
    "boys' choir": "voice",
    "girls' choir": "voice",
    "womens' choir": "voice",
    "mens' choir": "voice",
    "violin": "strings",
    "violin concertino": "strings",
    "violin principale": "strings",
    "violin ripieno": "strings",
    "solo violin part": "strings",
    "viola": "strings",
    "alto viola": "strings",
    "violoncello": "strings",
    "cello": "strings",
    "contrabass": "strings",
    "double bass": "strings",
    "": "keyboard",
    "baryton": "strings",
    "fiddle": "strings",
    "violetta": "strings",
    "violetta marina": "strings",
    "viola basso": "strings",
    "viola bastarda": "strings",
    "viola d'amore": "strings",
    "viola da braccio": "strings",
    "viola da gamba": "strings",
    "treble viola da gamba": "strings",
    "tenor viola da gamba": "strings",
    "bass viola da gamba": "strings",
    "viola pomposa": "strings",
    "violone": "strings",
    "strings": "general",
    "flute": "woodwinds",
    "transverse flute": "woodwinds",
    "piccolo": "woodwinds",
    "ottavino": "woodwinds",
    "flûte d'amour": "woodwinds",
    "flauto dolce": "woodwinds",
    "flauto quarto": "woodwinds",
    "flute in e|b": "woodwinds",
    "flautino": "woodwinds",
    "oboe": "woodwinds",
    "piccolo oboe": "woodwinds",
    "oboe grande": "woodwinds",
    "oboe d'amore": "woodwinds",
    "a ob": "woodwinds",
    "cor anglais": "woodwinds",
    "english horn": "woodwinds",
    "oboe da caccia": "woodwinds",
    "oboe di silva": "woodwinds",
    "taille de hautbois": "woodwinds",
    "vox humana": "woodwinds",
    "tenor oboe": "woodwinds",
    "heckelphone": "woodwinds",
    "b ob": "woodwinds",
    "clarinet": "woodwinds",
    "bass clarinet": "woodwinds",
    "clarinet in a": "woodwinds",
    "clarinet in c": "woodwinds",
    "e|b clarinet": "woodwinds",
    "basset horn": "woodwinds",
    "chalumeau": "woodwinds",
    "saxophone": "woodwinds",
    "soprano saxophone": "woodwinds",
    "alto saxophone": "woodwinds",
    "tenor saxophone": "woodwinds",
    "baritone saxophone": "woodwinds",
    "bassoon": "woodwinds",
    "contrabassoon": "woodwinds",
    "double bassoon": "woodwinds",
    "tenor bassoon": "woodwinds",
    "tenoron": "woodwinds",
    "bombardon": "brass",
    "bombarde": "woodwinds",
    "pommer": "woodwinds",
    "chirimía": "woodwinds",
    "shawm": "woodwinds",
    "fife": "woodwinds",
    "ocarina": "woodwinds",
    "recorder": "woodwinds",
    "sopranino recorder": "woodwinds",
    "woodwinds": "woodwinds",
    "zufoletto": "woodwinds",
    "wind band": "brass",
    "usually brass": "brass",
    "althorn": "brass",
    "also for unspecified bass instrument": "brass",
    "baritone horn": "brass",
    "flicorno": "brass",
    "bombardino": "brass",
    "bass tuba": "brass",
    "brass instruments": "general",
    "bugle": "brass",
    "cimbasso": "brass",
    "clarino": "brass",
    "clairon": "brass",
    "clarion": "brass",
    "cornetto": "brass",
    "cornett": "brass",
    "zink": "brass",
    "horn": "brass",
    "french horn": "brass",
    "keyed bugle": "brass",
    "bass horn": "brass",
    "valve horn": "brass",
    "corno da caccia": "brass",
    "dugetto": "brass",
    "euphonium": "brass",
    "fanfare trumpet": "brass",
    "flugelhorn": "brass",
    "basse d'harmonie": "brass",
    "helicon": "brass",
    "cornettino": "brass",
    "cornet": "brass",
    "lituus": "brass",
    "ophicleide": "brass",
    "piston": "brass",
    "rotary valve cornet": "brass",
    "post horn": "brass",
    "principale": "brass",
    "saxhorn": "brass",
    "serpent": "brass",
    "mellophone": "brass",
    "tuba": "brass",
    "tenor horn": "brass",
    "e|b horn": "brass",
    "trumpet": "brass",
    "keyed trumpet": "brass",
    "valve trumpet": "brass",
    "trombone": "brass",
    "alto trombone": "brass",
    "baritone trombone": "brass",
    "tenor trombone": "brass",
    "bass trombone": "brass",
    "valve trombone": "brass",
    "a small trumpet or tenor trombone": "brass",
    "a high trumpet": "brass",
    "archlute": "plucked",
    "harp": "plucked",
    "guitar": "plucked",
    "electric guitar": "plucked",
    "bass electric guitar": "plucked",
    "lyre": "plucked",
    "lute": "plucked",
    "mandolin": "plucked",
    "plucked instrument": "plucked",
    "requinto": "plucked",
    "psaltery": "plucked",
    "theorbo": "plucked",
    "zither": "plucked",
    "drums": "percussion",
    "birds singing": "percussion",
    "bell": "percussion",
    "handbell": "percussion",
    "chimes": "percussion",
    "small bell": "percussion",
    "glock": "percussion",
    "bells": "percussion",
    "carillon": "percussion",
    "tenor drum": "percussion",
    "parade drum": "percussion",
    "castanets": "percussion",
    "finger cymbals": "percussion",
    "dulcimer": "percussion",
    "cimbalom": "percussion",
    "claves": "percussion",
    "congas": "percussion",
    "conga drums": "percussion",
    "antique cymbals": "percussion",
    "glass harmonica": "percussion",
    "gong": "percussion",
    "bass drum": "percussion",
    "cuckoo": "percussion",
    "bell lyre": "percussion",
    "maracas": "percussion",
    "marimba": "percussion",
    "musical clock": "percussion",
    "musical glasses": "percussion",
    "rattle": "percussion",
    "nightingale": "percussion",
    "cymbal": "percussion",
    "chinese cymbal": "percussion",
    "suspended cymbal": "percussion",
    "ratchet": "percussion",
    "sleigh bells": "percussion",
    "spoons": "percussion",
    "percussion": "percussion",
    "tam tam": "percussion",
    "drum": "percussion",
    "string drum": "percussion",
    "tabor": "percussion",
    "drum with snares off": "percussion",
    "turkish drum": "percussion",
    "tambourine": "percussion",
    "snare drum": "percussion",
    "timbales": "percussion",
    "timpani": "percussion",
    "tom tom": "percussion",
    "triangle": "percussion",
    "vibraphone": "percussion",
    "xylophone": "percussion",
    "celesta": "keyboard",
    "cembalo": "keyboard",
    "clavicembalo": "keyboard",
    "spinet": "keyboard",
    "virginal": "keyboard",
    "etc. four hands": "keyboard",
    "clavecin": "keyboard",
    "clavichord": "keyboard",
    "harmonium": "keyboard",
    "harpsichord": "keyboard",
    "organ": "keyboard",
    "organ four hands": "keyboard",
    "barrel organ": "keyboard",
    "organo a rullo": "keyboard",
    "piano": "keyboard",
    "piano four hands": "keyboard",
    "physharmonica": "keyboard",
    "thoroughbass": "strings",
    "basso continuo": "strings",
    "english": "general",
    "continuous solo instrument in a work": "general",
    "winds": "general",
    "orchestra": "general",
    "additions to any wind instrument to indicate a tuning other than standard tuning": "general",
    "examples": "general",
    "electric": "electric",
    "cl in a \t\t\tcor in e|b": "general",
}
"""Dictionary containing the family of each instrument (sound)"""

SCORING_ORDER = [
    "fl",
    "ob",
    "eh",
    "cl",
    "bn",
    "hn",
    "tpt",
    "tb",
    "clno",
    "timp",
    "tamb",
    "triang",
    "mand",
    "salt",
    "vv",
    "cv",
    "eq",
    "fch",
    "sop",
    "alt",
    "ten",
    "bar",
    "bbar",
    "bas",
    "vn",
    "va",
    "viol",
    "vc",
    "bs",
]
"""List of sounds/instruments in the order we want them TODO: better documentation here!"""

SCORING_FAMILY_ORDER = ["ww", "br", "perc", "pluc", "voice", "str"]
"""List of families in the order we want them TODO: better documentation here!"""
